from kensu.utils.wrappers import remove_ksu_wrappers, remove_ksu_kwargs_wrappers

from kensu.utils.kensu_provider import KensuProvider


def wrap_pandas_gbq_write(method):
    def wrapper(*args, **kwargs):
        kensu = KensuProvider().instance()

        new_args = remove_ksu_wrappers(args)
        new_kwargs = remove_ksu_kwargs_wrappers(kwargs)

        df_result = method(*new_args, **new_kwargs)
        df = args[0]  # see get_dummies definition (first arg is `data`)

        orig_ds = kensu.extractors.extract_data_source(df, kensu.default_physical_location_ref,
                                                     logical_data_source_naming_strategy=kensu.logical_naming)
        orig_sc = kensu.extractors.extract_schema(orig_ds, df)

        from google.cloud import bigquery
        client = bigquery.Client()
        df_table = client.get_table(args[1])

        result_ds = kensu.extractors.extract_data_source(df_table, kensu.default_physical_location_ref,
                                                       logical_data_source_naming_strategy=kensu.logical_naming)._report()
        result_sc = kensu.extractors.extract_schema(result_ds, df_table)._report()


        for col in df.columns:
            kensu.add_dependencies_mapping(result_sc.to_guid(), str(col), orig_sc.to_guid(), str(col), 'Write')

        kensu.real_schema_df[result_sc.to_guid()] = df
        kensu.report_with_mapping()

        return df_result

    wrapper.__doc__ = method.__doc__
    return wrapper


import os

# P.S. this file should use py3.5 syntax, no f-strings allowed

def extract_config_property(key, default, arg=None, kw=None, conf=None, tpe=None):
    """
    Looks for a property value following this precedence:
      env_var > arg > kwargs > conf > default
    The default value is used to determine the type of the conf value (it can be overridden by tpe).
    The environment variable will be looked up based on the pattern of `KSU_<upper-key>`.
    """
    env_var_key = "KSU_" + key.upper()
    if os.environ.get(env_var_key) is not None:
        env_var = os.environ.get(env_var_key)
        if tpe is not None:
            env_var = tpe(env_var)
        return env_var
    elif arg is not None:
        return arg
    elif key in kw and kw[key] is not None:
        return kw[key]
    elif key in conf and conf.get(key) is not None:
        if default is not None and tpe is None:
            tpe = type(default)
        r = conf.get(key)
        if tpe is list:
            r = r.replace(" ", "").split(",")
        elif tpe is bool:
            r = conf.getboolean(key)
        elif tpe is not None:
            r = tpe(r)
        return r
    else:
        return default


def get_conf_path(default = "conf.ini"):
    return os.environ["KSU_CONF_FILE"] if "KSU_CONF_FILE" in os.environ else default



import logging
import os
from datetime import datetime
import re

import requests

from abc import ABC, abstractmethod

# we could reuse the reporters instead
from kensu.utils.exceptions import SdkError


class AbstractSDK(ABC):

    @abstractmethod
    def get_lineages_in_project(self, project, process, env, code_version):
        pass

    @abstractmethod
    def create_rule(self, lds_id, lineage_id, project_id, process_id, env_name, field_name, fun):
        pass

    @abstractmethod
    def update_rule(self, predicate, fun):
        pass

    @abstractmethod
    def get_rules(self):
        pass

    @abstractmethod
    def get_all_rules_for_ds(self, ds_id):
        pass

    @abstractmethod
    def get_rules_for_ds_in_project(self, ds_id, lineage_id, project_id, env):
        pass

    @abstractmethod
    def get_logical_datasources(self):
        pass

    @abstractmethod
    def get_total_sent_data_mb(self):
        pass

    @abstractmethod
    def get_detailed_sent_data(self):
        pass

    @abstractmethod
    def disable_all_metrics(self, ds_reg_ex):
        pass

    @abstractmethod
    def set_metrics(self, ds_reg_ex, attributes_reg_ex, whitelist):
        pass

    @abstractmethod
    def get_tickets(self, status):
        pass

    @abstractmethod
    def set_ticket_status(self, id):
        pass


def normalize_services_response(func):
    def wrapper(*args, **kwargs):
        res = func(*args, **kwargs)
        if res and 'data' not in res:
            return {'data': res}
        return res

    wrapper.__doc__ = func.__doc__
    return wrapper


class DoNothingSDK(ABC):
    def __init__(self):
        pass

    def get_lineages_in_project(self, project, process, env, code_version):
        pass

    def create_rule(self, lds_id, lineage_id, project_id, process_id, env_name, field_name, fun):
        pass

    def update_rule(self, predicate, fun):
        pass

    def get_rules(self):
        pass

    def get_all_rules_for_ds(self, ds_id):
        pass

    def get_rules_for_ds_in_project(self, ds_id, lineage_id, project_id, env):
        pass

    def get_logical_datasources(self):
        pass

    def get_total_sent_data_mb(self):
        pass

    def get_detailed_sent_data(self):
        pass

    def disable_all_metrics(self, ds_reg_ex):
        pass

    def get_tickets(self, status):
        pass

    def set_ticket_status(self, id):
        pass

    def set_metrics(self, ds_reg_ex, attributes_reg_ex, whitelist):
        pass

class SDK(AbstractSDK):
    def __init__(self, url, sdk_token, verify_ssl):
        self.url = url
        self.PAT = sdk_token
        self.verify_ssl = verify_ssl
        self.debug_requests = os.environ.get('KENSU_DEBUG_HTTP_REQUESTS', 'False') == 'True'

    def _auth_headers(self):
        return {"Authorization": f"Bearer {self.PAT}"}

    def requests_get_json(self, uri_suffix):
        # FIXME: proper URI concat
        uri = self.url + uri_suffix
        resp = self.debug_request(
            uri=uri,
            method='GET',
            fn=lambda: requests.get(uri, headers=self._auth_headers(), verify=self.verify_ssl)
        )
        if not (200 <= resp.status_code <= 299):
            msg = f"Failed to query Kensu SDK for uri={uri_suffix} status={resp.status_code}:\n{resp.text}"
            e = SdkError(msg)
            logging.warning(msg, e)
            raise SdkError(e)
        try:
            return resp.json()
        except:
            logging.warning(f"Unable to decode Kensu SDK response for uri={uri_suffix}:\n{v}", e)
            raise SdkError()

    def debug_request(self, uri, method, fn, payload=None):
        if self.debug_requests:
            logging.warning(f'performing {method} request to {uri}, payload: {payload}')
        resp = fn()
        if self.debug_requests:
            logging.warning('got response: ' + str(resp))
            logging.warning('got response text: ' + str(resp.text))
        return resp

    def requests_post_json(self, uri, payload):
        full_uri = self.url + uri
        return self.debug_request(
            uri=full_uri,
            payload=payload,
            method='POST',
            fn=lambda: requests.post(full_uri, headers=self._auth_headers(), json=payload, verify=self.verify_ssl)
        )

    def requests_put_json(self, uri, payload):
        full_uri = self.url + uri
        return self.debug_request(
            uri=full_uri,
            payload=payload,
            method='PUT',
            fn=lambda: requests.put(full_uri, headers=self._auth_headers(), json=payload, verify=self.verify_ssl)
        )

    def requests_patch_json(self, uri, payload):
        full_uri = self.url + uri
        return self.debug_request(
            uri=full_uri,
            payload=payload,
            method='PATCH',
            fn=lambda: requests.patch(full_uri, headers=self._auth_headers(), json=payload, verify=self.verify_ssl)
        )

    def get_lineages_in_project(self, project, process, env, code_version):
        # FIXME: this works only when explicit environment was specified (and maybe same about code version)
        # FIXME: use proper URLencode
        uri = "/business/services/views/v1/project-catalog/process/data-flow?projectId=%s&processId=%s&environment=%s&codeVersionId=%s" % (
            project, process, env, code_version)
        return {'data': self.requests_get_json(uri)}

    def get_business_services_version(self):
        uri = '/business/services/v1/code-version'
        return self.requests_get_json(uri)['version']

    def create_rule(self, lds_id, lineage_id=None, project_id=None, process_id=None, env_name=None, field_name=None,
                    fun=None, context="DATA_STATS"):
        uri = "/business/services/v1/rules"
        payload = None
        if context == "DATA_STATS":
            payload = {
                "context": "DATA_STATS",
                "datasourceId": lds_id,
                "lineageId": lineage_id,
                "projectId": project_id,
                "processId": process_id,
                "environment": env_name,
                "isLogical": True,
                "fieldName": field_name,
                "functionName": fun["name"],
                "arguments": fun["arguments"]
            }
        elif context == "LOGICAL_DATA_SOURCE":
            payload = {
                "context": "LOGICAL_DATA_SOURCE",
                "datasourceId": lds_id,
                "environment": "",
                "fieldName": field_name,
                "functionName": fun["name"],
                "arguments": fun["arguments"]
            }

        if payload:
            # FIXME: verify no failure?
            return self.requests_post_json(uri, payload=payload)

    def update_rule(self, predicate, fun):
        payload = {"functionName": fun["name"],
                   "arguments": fun["arguments"]}
        # e.g.: /business/services/v1/rules/1d7054fd-5f0c-4c71-b94f-2f0fe8deb210
        uri = f"/business/services/v1/rules/{predicate}"
        v = self.requests_patch_json(uri, payload=payload)
        return None

    @normalize_services_response
    def get_rules(self):
        # FIXME: not used?
        uri = "/business/services/views/v1/rules"
        return self.requests_get_json(uri)

    @normalize_services_response
    def get_rules_for_ds_in_project(self, ds_id, lineage_id, project_id, env):
        uri = f"/business/services/views/v2/performance/data/{ds_id}/{lineage_id}?projectId={project_id}&environment={env}"
        return self.requests_get_json(uri)

    def get_all_rules_for_ds(self, ds_id):
        uri = "/business/services/views/v1/rules?logical_data_source_id=%s&context=LOGICAL_DATA_SOURCE" % (ds_id)
        r = self.requests_get_json(uri)
        return {'data': {'predicates': r}}

    def get_logical_datasources(self):
        logical_data_sources = {'data': []}
        limit = 500
        offset = 0

        ds = {'data': [None]} 
    
        while ds['data']:
            try:
                ds = self.requests_get_json(f"/business/services/views/v1/logical-datasources?offset={offset}&limit={limit}")
                logical_data_sources['data'].extend(ds['data'])
                offset += limit
                
            except Exception as e:
                print(f"Error fetching data: {e}")
                break # Exit loop on error
   
        return logical_data_sources

    def get_total_sent_data_mb(self):
        return self.requests_get_json("/business/services/v1/ingestion-log")['entitiesSentKbTotal'] / 1000

    def get_detailed_sent_data(self):
        ingestion_log_details = self.requests_get_json("/business/services/v1/ingestion-log")['entitiesSentByToken']
        tokens = self.requests_get_json("/business/services/v1/tokens/app-groups")
        ingestion_list = []
        for ingestion_log_detail in ingestion_log_details:
            ingestion_dict = {}
            if ingestion_log_detail['entitiesSent'] > 0:
                token = next(token for token in tokens if token['id'] == ingestion_log_detail['ingestionTokenId'])
                ingestion_dict = {
                    'week': datetime.fromtimestamp(ingestion_log_detail['ingestedAt'] / 1000).strftime('%Y/%V'),
                    'sizeMb': ingestion_log_detail['entitiesSentKbSize'] / 1000,
                    'entities': ingestion_log_detail['entitiesSent'],
                    'tokenId': token['id'],
                    'tokenDescription': token['tokenDescription'],
                    'appGroup': token['appGroupName']
                }
                ingestion_list.append(ingestion_dict)
        return ingestion_list

    def disable_all_metrics(self, ds_reg_ex):
        dss = self.get_logical_datasources()
        for ds in dss:
            x = re.search(ds_reg_ex, ds['name'])  # check if ds name match with the regular exp
            if x:
                apps = self.requests_get_json(  # get list of all applications involved
                    "/business/services/views/v2/performance/stats/attributes?logicalDataSourceId=%s" % ds['id'])[
                    'applications']
                for app in apps:
                    logging.info("Disable metrics for " + ds['name'] + " and application " + app['name'])
                    self.requests_patch_json(
                        "/business/services/views/v1/logical-datasources/" + ds['id'] + "/configuration/processes/" +
                        app['id'] + "/metrics",
                        {"isMetricsConfigurationEnabled": True, "activeMetrics": [], "circuitBreakerRuleIds": []})

    def set_metrics(self, ds_reg_ex, attributes_reg_ex, whitelist=True):
        dss = self.get_logical_datasources()
        for ds in dss:
            x = re.search(ds_reg_ex, ds['name'])  # check if ds name match with the regular exp
            if x:
                apps = self.requests_get_json(  # get list of all applications involved
                    "/business/services/views/v2/performance/stats/attributes?logicalDataSourceId=%s" % ds['id'])[
                    'applications']
                schema = self.requests_get_json(  # get list of all the possible metrics
                    "/business/services/views/v1/logical-datasources/%s/schemas" % ds['id'])
                attributes = []
                for field in schema:
                    attributes.append(field['columnName'] + ".nullrows")
                    attributes.append(field['columnName'] + ".nrows")
                    if field['metricType'] == "NUMERIC":
                        attributes.append(field['columnName'] + ".min")
                        attributes.append(field['columnName'] + ".max")
                        attributes.append(field['columnName'] + ".mean")
                        attributes.append(field['columnName'] + ".stddev")
                    if field['metricType'] == "TIMESTAMP":
                        attributes.append(field['columnName'] + ".first")
                        attributes.append(field['columnName'] + ".last")
                filter_attributes = re.compile(attributes_reg_ex)
                if whitelist:  # whitelisting
                    attributes = list(filter(filter_attributes.search, attributes))
                else:  # blacklisting
                    attributes = list(filter(lambda x: not filter_attributes.search(x), attributes))
                attributes.append('nrows')
                for app in apps:
                    logging.info("Enable specified metrics for " + ds['name'] + " and application " + app['name'])
                    # CB rules are preserved
                    payload = {"isMetricsConfigurationEnabled": True, "activeMetrics": attributes}
                    self.requests_patch_json(
                        "/business/services/views/v1/logical-datasources/" + ds['id'] + "/configuration/processes/" +
                        app['id'] + "/metrics", payload)

    def get_tickets(self, status='OPEN'):
        limit = 20
        payload = {
            "limit": limit,
            "status": status,
            "sortBy": "CREATION_TIMESTAMP",
            "order": "DESC",
            "offset": 0
        }
        result = self.requests_post_json("/business/services/views/v1/tickets", payload=payload)
        tickets = result.json()['data']
        pointer = result.json()['pagination']
        while pointer['hasNext']:
            payload['offset'] = payload['offset'] + limit
            result = self.requests_post_json("/business/services/views/v1/tickets", payload=payload)
            tickets.extend(result.json()['data'])
            pointer = result.json()['pagination']
        return tickets

    def set_ticket_status(self, id):
        payload = {
            "resolution":"Resolved","status":"CLOSED","ticketIds":[id]
        }
        return self.requests_put_json("/business/services/views/v1/tickets/status", payload=payload)



import logging
from typing import Dict

from airflow.operators import bash as airflow_bash
from airflow.operators.bash import *

from kensu.airflow.kensu_airflow_collector import COLLECTOR_STATUS_INIT, COLLECTOR_STATUS_DONE, log_status, \
    airflow_init_kensu, handle_ex
from kensu.utils.helpers import report_simple_copy_with_guessed_schema, get_absolute_path


def dumb_parse_curl(s):
    import re
    try:
        # This is just a very simple example and should be customized to customer needs
        # e.g.: curl -sSL https://s3.amazonaws.com/nyc-tlc/trip+data/yellow_tripdata_2021-01.csv > /home/airflow/gcs/data/yellow_tripdata_2021-01.csv
        regexp = re.compile(r'curl.* (http[^> ]+) > (.+)')
        logging.info(f"dumb_parse_curl(s={s})")
        m = regexp.match(s)
        if m:
            logging.info("dumb_parse_curl parsed groups: " + str(m.groups()))
            input_uri = m.group(1)
            output_filename = m.group(2)
            output_absolute_uri = get_absolute_path(output_filename)
            report_simple_copy_with_guessed_schema(
                input_uri=input_uri,
                output_absolute_uri=output_absolute_uri,
                read_schema_from_filename=output_filename,
                operation_type='airflow.BashOperator::curl'
            )
            return True
    except Exception as e:
        logging.warning(f"caught exception in dumb_parse_curl", e)

    return None


class BashOperator(airflow_bash.BashOperator):
    def __init__(
            self,
            *args,
            **kwargs,
    ):
        super(BashOperator, self).__init__(
            *args, **kwargs
        )
        self.cmd_matchers = [
            dumb_parse_curl
        ]

    def execute(self, context: Dict) -> None:
        res = super(BashOperator, self).execute(context)
        try:
            log_status(self, COLLECTOR_STATUS_INIT)
            airflow_init_kensu(airflow_operator=self)
            for matcher in self.cmd_matchers:
                if matcher(self.bash_command):
                    break
        except Exception as ex:
            handle_ex(self, ex)
        log_status(self, COLLECTOR_STATUS_DONE)
        return res


BashOperator.__doc__ = airflow_bash.BashOperator.__doc__



# this file should not use f-strings syntax
# because f-strings are not supported in python 3.5
# so instead of: f"my formatted {text}" # py3.6+
# we use: "my formatted {}".format(text) # py3.5+

import logging


def raise_default_circuit_breaker_exception(msg):
    raise AssertionError(msg)


def kill_app_on_circuit_breaker(
        exit_code=254,
        spark=None,
        stop_spark=True,
        sys_exit=True,
        raise_exception_fn=lambda msg: raise_default_circuit_breaker_exception(msg)
):
    if spark is not None:
        logging.info('Going to check the circuit breakers for Spark datasources')
        from kensu.pyspark.spark_connector import check_spark_circuit_breakers_and_stop_if_broken
        breakers_failed = check_spark_circuit_breakers_and_stop_if_broken(spark=spark, stop_spark=stop_spark)

        if breakers_failed:
            if sys_exit:
                import sys
                logging.warning("Some kensu circuit breaker has failed - killing the app with exit_code={}".format(exit_code))
                sys.exit(exit_code)
            else:
                msg = "Some kensu circuit breaker has failed - raising an Exception to stop the default app flow"
                logging.warning(msg)
                raise_exception_fn(msg)
                return False
        else:
            logging.info("Circuit breakers did not fail")
        return breakers_failed
    else:
        logging.info('Not checking any Circuit breakers (spark=None)')
        return True


"""
To be used interactively, as it do not kill spark and do not call sys.exit(), just raises an exception
"""
def databricks_notebook_break_app_on_circuit_breaker(
        spark=None,
        raise_exception_fn=lambda msg: raise_default_circuit_breaker_exception(msg)
):
    breakers_failed = kill_app_on_circuit_breaker(
        spark=spark,
        stop_spark=False,  # FIXME: not clear if better to stop
        sys_exit=False,
        raise_exception_fn=raise_exception_fn
    )
    return breakers_failed



import gluonts.model.forecast as spf
from kensu.utils.helpers import eventually_report_in_mem
from kensu.utils.kensu_provider import KensuProvider


class SampleForecast(spf.SampleForecast):
    dependencies = None

    def create_deps(self, result):
        deps = self.dependencies

        kensu = KensuProvider().instance()
        for orig_sc in deps:

            result_ds = eventually_report_in_mem(
                kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref,
                                                     logical_data_source_naming_strategy=kensu.logical_naming))
            result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, result))

            for result_col in [k.name for k in result_sc.pk.fields]:
                for orig_col in [k.name for k in orig_sc.pk.fields]:
                    kensu.add_dependencies_mapping(result_sc.to_guid(), str(result_col), orig_sc.to_guid(), str(orig_col),
                                                   "Model")



    @property
    def mean(self):
        import numpy as np
        from kensu.numpy import ndarray
        if self._mean is not None:
            return self._mean
        else:
            result = ndarray.using(np.mean(self.samples, axis=0))
            self.create_deps(result)
            return result


import logging
import re
from typing import Dict, Iterator, NamedTuple, Optional, Tuple

# Third-party imports

import gluonts  # noqa
from gluonts import transform
from gluonts.core.serde import load_code
from gluonts.dataset.common import DataEntry, Dataset
from gluonts.dataset.stat import (
    DatasetStatistics,
    calculate_dataset_statistics,
)
from gluonts.evaluation import Evaluator
from gluonts.model.estimator import Estimator
from gluonts.model.forecast import Forecast
from gluonts.model.predictor import Predictor
from gluonts.support.util import maybe_len
from gluonts.transform import TransformedDataset
import pandas as pd
from kensu.gluonts.ksu_utils.dataset_helpers import make_dataset_reliable


def make_evaluation_predictions(
        dataset: Dataset, predictor: Predictor, num_samples: int
) -> Tuple[Iterator[Forecast], Iterator[pd.Series]]:
    import pandas as pd

    prediction_length = predictor.prediction_length
    freq = predictor.freq
    lead_time = predictor.lead_time

    def add_ts_dataframe(
            data_iterator: Iterator[DataEntry],
    ) -> Iterator[DataEntry]:
        for data_entry in data_iterator:
            data = data_entry.copy()
            index = pd.date_range(
                start=data["start"],
                freq=freq,
                periods=data["target"].shape[-1],
            )
            data["ts"] = pd.DataFrame(
                index=index, data=data["target"].transpose()
            )
            yield data

    def ts_iter(dataset: Dataset) -> pd.DataFrame:
        dataset, old, dep_fields = make_dataset_reliable(dataset)
        for data_entry in add_ts_dataframe(iter(dataset)):
            yield data_entry["ts"]
        dataset.list_data = old

    def truncate_target(data):
        data = data.copy()
        target = data["target"]
        assert (
                target.shape[-1] >= prediction_length
        )  # handles multivariate case (target_dim, history_length)
        data["target"] = target[..., : -prediction_length - lead_time]
        return data

    # TODO filter out time series with target shorter than prediction length
    # TODO or fix the evaluator so it supports missing values instead (all
    # TODO the test set may be gone otherwise with such a filtering)

    dataset_trunc = TransformedDataset(
        dataset, transformation=transform.AdhocTransform(truncate_target)
    )

    return (predictor.predict(dataset_trunc, num_samples=num_samples), ts_iter(dataset))


from kensu.pandas import DataFrame
from kensu.utils.kensu_provider import KensuProvider
import sklearn.neighbors as ngb
from kensu.utils.helpers import eventually_report_in_mem


class KNeighborsClassifier(ngb.KNeighborsClassifier):

    def fit(self, X, y):
        X_train = X.get_df()
        y_train = y.get_s()
        result = super(KNeighborsClassifier,self).fit( X_train , y_train )
        kensu = KensuProvider().instance()

        metrics = kensu.extractors.extract_machine_learning_metrics(result,x_train = X_train, y_train = y_train )
        kensu = KensuProvider().instance()

        train_X_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(X_train, kensu.default_physical_location_ref))
        train_X_sc = eventually_report_in_mem(kensu.extractors.extract_schema(train_X_ds, X_train))

        train_y_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(y, kensu.default_physical_location_ref))
        train_y_sc = eventually_report_in_mem(kensu.extractors.extract_schema(train_y_ds, y))

        result_ds = eventually_report_in_mem(kensu.extractors.extract_data_source(result, kensu.default_physical_location_ref,format='Sklearn Model'))
        result_sc = eventually_report_in_mem(kensu.extractors.extract_schema(result_ds, result))

        hyperparams = result.get_params()

        if kensu.mapping:
            for result_col in [s.name for s in result_sc.pk.fields]:
                for col in [s.name for s in train_X_sc.pk.fields]:
                    kensu.add_dependencies_mapping(result_sc.to_guid(),result_col,train_X_sc.to_guid(),col,'Model Train')
                for col in [s.name for s in train_y_sc.pk.fields]:
                    kensu.add_dependencies_mapping(result_sc.to_guid(), result_col, train_y_sc.to_guid(), col,
                                                 'Model Train')

        attr = [result,'SkLearn.KNeighborsClassifier',metrics,hyperparams]
        self.attr = attr

        return result


import logging

import urllib3

urllib3.disable_warnings()
from kensu.utils.kensu import Kensu
from kensu.client.models import *
from kensu.utils.helpers import to_datasource

from pyspark.sql import DataFrame
from pyspark.sql import SparkSession
from kensu.pyspark import get_process_run_info, get_inputs_lineage_fn, register_manual_lineage, get_spark_from_df, \
    get_df_with_inmem_tag
from kensu.utils.kensu_provider import KensuProvider


def get_spark_session():
    kensu_spark_session = None
    try:
        import gc
        for obj in gc.get_objects():
            if isinstance(obj, SparkSession):
                kensu_spark_session = obj
    except:
        return None
    else:
        return kensu_spark_session


import types


def get_schema_fields_from_spark_df(
        spark_df,  # type: DataFrame
):
    return list([
        FieldDef(name=f.name, field_type=f.dataType.simpleString(), nullable=False)
        for f in spark_df.schema.fields
    ])


def register_in_mem_schema(
        ds,  # type: DataSource
        schem_fields,  # type: list[FieldDef]
        in_mem_location,  # type: str
        in_mem_name  # type: str
):
    schema = Schema(in_mem_location, pk=SchemaPK(data_source_ref=DataSourceRef(by_guid=ds.to_guid()),
                                                 fields=schem_fields))._report()
    k = KensuProvider().instance()
    k.name_schema_lineage_dict[in_mem_name] = schema.to_guid()
    return schema


"""
Tag a Spark DataFrame as in-mem datasource, assuming that it will be transformed into python by using say .toPandas().
The result in Kensu will be: [spark input{1,2,..}] --lineage--> [in-mem: name]

Currently only a single .tagInMem() call is allowed on same Spark DataFrame. 
The subsequent calls would be interpreted as individual in-mem data-sources not dependent on the earlier one (due to
how spark works - DataFrame is never materialized before "write"), i.e.:
    `spark_df.tagInMem('name1').filter('col1 = 1').tagInMem('name2')`
would create two UNRELATED in-mem data-sources.
"""


def tagInMem(self,  # type: DataFrame
             name):
    try:
        # Get spark session info
        # FIXME: we can get Spark session more reliably/efficiently from `self: DataFrame`,
        #  while databricks have multiple sessions (and not sure about isolation there)!
        # spark = get_spark_session()
        # FIXME: check if self is Spark DataFrame?
        spark_df = self
        spark = get_spark_from_df(spark_df)
        spark_info = get_process_run_info(spark)
        process = spark_info['process_guid']

        # Creation of an in-mem data source
        spark_ds_name = f'{process}/{name}'
        # location = f'in-memory-data://{process}/{name}'
        location = f'in-memory-data://{spark_ds_name}'
        format = 'in-mem'
        k = KensuProvider().instance()
        ds_pk = DataSourcePK(location=location,
                             physical_location_ref=PhysicalLocationRef(by_pk=k.UNKNOWN_PHYSICAL_LOCATION.pk))
        df_ds = to_datasource(ds_pk, format, location, 'File', name)._report()
        spark_df_schema = register_in_mem_schema(
            ds=df_ds,
            schem_fields=get_schema_fields_from_spark_df(spark_df),
            in_mem_location=location,
            in_mem_name=name
        )
        spark_df_schema._report()

        # report Spark lineage to Kensu
        # FIXME: check if self is Spark DataFrame
        # report lineage to Kensu
        spark_lineage = get_inputs_lineage_fn(kensu_instance=k, df=spark_df, virtual_ds_name=spark_ds_name)
        print('got lineage from spark, just before calling .toPandas()')
        print(spark_lineage)
        # FIXME: we could use this and also report spark datastats
        # FIXME: or should we report using low level apis instead, so we'd easily keep the process info extracted from spark
        # spark_lineage.report(
        #     ksu=k,
        #     df_result=spark_df_schema,  # we don't handle Spark DataFrame directly in extractors yet, so pass a schema
        #     operation_type='Spark to Python',
        #     report_output=True,
        #     register_output_orig_data=False
        # )
        # k.report_with_mapping()
        report_spark_lineage(spark_lineage, k=k, result_schema=spark_df_schema, output_name=name)
    except Exception as e:
        logging.warning(f"tagInMem for toPandas didnt succeed to extract lineage within given time limit", e)

    return self


def report_spark_lineage(spark_lineage,
                         k,  # type: Kensu
                         result_schema,
                         output_name):
    # similar like GenericComputedInMemDs.report, but using processRun from spark and no Kensu-Py magic
    from kensu.utils.helpers import extract_ksu_ds_schema
    spark_input_names = []
    for input_ds in spark_lineage.inputs:
        ds, schema = extract_ksu_ds_schema(k, input_ds, report=True, register_orig_data=False)
        input_name = ds.pk.location
        spark_input_names.append(input_name)
        k.name_schema_lineage_dict[input_name] = schema.to_guid()

    # report output (if needed)
    lineage_run = link(input_names=spark_input_names, output_name=output_name)
    lineage_run_id = lineage_run.to_guid()
    # ask spark to publish stats given a lineage-run
    # FIXME: do we need to activate/deactivate it depending on config? but kensupy conf or pyspark conf?
    for input_ds in spark_lineage.inputs:
        input_ds.f_publish_stats(lineage_run_id)


def tagInMemWrapper():
    def tagInMemInner(self,  # type: DataFrame
                      name  # type: str
                      ):
        return tagInMem(self, name)

    return tagInMemInner


def tagCreateSparkDataFrameFromPy(
        self,  # type: DataFrame
        name,  # type: str
        input_names=None  # type: list[str]
):
    try:
        # Get spark session info
        spark = get_spark_from_df(self)
        spark_info = get_process_run_info(spark)
        process = spark_info['process_guid']

        # Creation of an in-mem data source
        location = f'in-memory-data://{process}/{name}'
        in_mem_format = 'in-mem'
        k = KensuProvider().instance()
        ds_pk = DataSourcePK(location=location,
                             physical_location_ref=PhysicalLocationRef(by_pk=k.UNKNOWN_PHYSICAL_LOCATION.pk))
        ds = to_datasource(ds_pk, in_mem_format, location, 'File', name)._report()

        # FIXME: only spark dataframe supported now, i.e. this function must be after spark.createDataFrame()
        fields = get_schema_fields_from_spark_df(spark_df=self)
        schema = Schema(location,
                        pk=SchemaPK(data_source_ref=DataSourceRef(by_guid=ds.to_guid()), fields=fields))._report()
        k.name_schema_lineage_dict[name] = schema.to_guid()

        # report lineage to Kensu between Python inputs and spark.createDataFrame which is not seen by Spark directly
        if input_names:
            link(input_names=input_names, output_name=name)
        else:
            # p.s. if we need mode where in-mem is not reported, would need to get earlier reported lineage
            pass

        # also need to report lineage to Spark JVM,
        # which for now will be simply `name` of this .createDataFrame() in-mem operation
        output_df_tag = f'spark_df_{name}'
        register_manual_lineage(
            spark,
            output_df_tag=output_df_tag,  # p.s. output_df_tag is only used as an ID in Spark agent side, nothing else
            # P.S. or alternatively we could pass over the input_names to Spark,
            #  and then not report any lineage between inputs_name(python) and {spark_dataframe} from within Python
            #  but then the `name` (of this in-memory spark dataframe)  would be not reported to Kensu, so
            #  current solution seems better/more consistent with the rest of this file
            inputs=[
                {
                    'path': location,  # FIXME: will DS name be consistently reported between python and Spark/scala?
                    'format': in_mem_format,
                    'schema': dict([(f.name, f.field_type) for f in fields])
                },
            ]
        )
        output_df = get_df_with_inmem_tag(df=self, df_tag=output_df_tag)
    except Exception as e:
        logging.warning(f"tagCreateSparkDataFrameFromPy encountered an issue", e)
        output_df = self
    return output_df


def tagCreateDataFrameWrapper():
    def tagInMemInner(self,  # type: DataFrame
                      name,  # type: str
                      input_names=None
                      ):
        return tagCreateSparkDataFrameFromPy(self, name=name, input_names=input_names)

    return tagInMemInner


def add_tagInMem_operations():
    DataFrame.tagInMem = tagInMemWrapper()
    DataFrame.tagCreateDataFrame = tagCreateDataFrameWrapper()


def create_publish_for_data_source(name, format, location, schema=None, **kwargs):
    k = KensuProvider().instance()
    ds_pk = DataSourcePK(location=location,
                         physical_location_ref=PhysicalLocationRef(by_pk=k.UNKNOWN_PHYSICAL_LOCATION.pk))

    ds = DataSource(name=name, format=format, categories=[f'logical::{name}'], pk=ds_pk)._report()

    if schema is not None:
        fields = schema
    else:
        fields = [FieldDef('unknown', 'unknown', True)]
    schema = Schema(name, pk=SchemaPK(data_source_ref=DataSourceRef(by_guid=ds.to_guid()), fields=fields))._report()
    k.name_schema_lineage_dict[name] = schema.to_guid()


def create_publish_for_postgres_table(table, location, cur=None):
    format = 'Postgres Table'
    from kensu.psycopg2.pghelpers import get_table_schema
    try:
        if cur:
            SchemaFields = get_table_schema(cur=cur, table_name=table)
            from kensu.client.models import FieldDef
            schema_ = [FieldDef(f['field_name'], f['field_type'], True) for f in SchemaFields]
        else:
            schema_ = None
    except:
        schema_ = None
    create_publish_for_data_source(ds=table, name=location, location=location, format=format, schema=schema_)


def create_publish_for_csv(name, location):
    format = 'csv'
    import pandas as pd
    from kensu.pandas.extractor import KensuPandasSupport
    schema_ = None
    create_publish_for_data_source(ds=name, name=name, location=location, format=format, schema=schema_)


def create_publish_for_sklearn_model(name, location):
    format = 'SKLearn'
    k = KensuProvider().instance()
    ds_pk = DataSourcePK(location=location,
                         physical_location_ref=PhysicalLocationRef(by_pk=k.UNKNOWN_PHYSICAL_LOCATION.pk))
    ds = DataSource(name=name, format=format, categories=[f'logical::{name}'], pk=ds_pk)._report()

    fields = [FieldDef('intercept', 'Numeric', False), FieldDef('coeff', 'Numeric', False)]
    schema = Schema(name, pk=SchemaPK(data_source_ref=DataSourceRef(by_guid=ds.to_guid()), fields=fields))._report()
    k.name_schema_lineage_dict[name] = schema.to_guid()


def get_schema(name):
    # use the lookup table

    k = KensuProvider().instance()
    if name in k.name_schema_lineage_dict:
        return k.name_schema_lineage_dict[name]
    else:
        logging.info(f'Unable to retrieve Kensu GUID for schema {name}')
        return None


def create_lineage(inputs, output):
    data_flow = [SchemaLineageDependencyDef(from_schema_ref=SchemaRef(by_guid=i),
                                            to_schema_ref=SchemaRef(by_guid=output),
                                            # FIXME: or should we report column_data_dependencies=None ?
                                            column_data_dependencies={'unknown': ['unknown']}) for i in inputs]

    # Definition of the process
    if get_spark_session() != None:
        process_guid = get_process_run_info(get_spark_session())['process_guid']
    else:
        k = KensuProvider().instance()
        process_guid = k.process.to_guid()

    lineage = ProcessLineage(name=f'Lineage to {str(output)}',  # FIXME: List inputs too?
                             operation_logic='APPEND',
                             pk=ProcessLineagePK(process_ref=ProcessRef(
                                 by_guid=process_guid),
                                                 data_flow=data_flow))._report()
    return lineage


def link(input_names, output_name):
    # retrieve schemas for inputs
    input_scs = [get_schema(i) for i in input_names]
    # retrieve schemas for outputs
    output_sc = get_schema(output_name)
    # create lineage
    lineage = create_lineage(input_scs, output_sc)._report()

    # Definition of the process run
    if get_spark_session() != None:
        process_run_guid = get_process_run_info(get_spark_session())['process_run_guid']
    else:
        k = KensuProvider().instance()
        process_run_guid = k.process_run.to_guid()

    # create lineage run
    k = KensuProvider().instance()
    lineage_run = LineageRun(LineageRunPK(process_run_ref=ProcessRunRef(by_guid=process_run_guid),
                                          lineage_ref=ProcessLineageRef(by_guid=lineage.to_guid()),
                                          # Should be int/long
                                          timestamp=int(k.timestamp)))
    print(f'reporting lineage_run for {input_names} -> {output_name}. payload={lineage_run}...')
    lineage_run._report()
    return lineage_run



import sys
from typing import overload, Dict, List, Optional, Tuple, TYPE_CHECKING, Union

from pyspark.sql.utils import to_str

from pyspark.sql.dataframe import DataFrameWriter

if TYPE_CHECKING:
    from pyspark.sql._typing import OptionalPrimitiveType
    from pyspark.sql.dataframe import DataFrame

__all__ = ["KensuDataFrameWriter"]

PathOrPaths = Union[str, List[str]]
TupleOrListOfString = Union[List[str], Tuple[str, ...]]


# limitation: partition columns are not supported for LDS name for remote config

# copied from: from pyspark.sql.utils import to_str
def to_str(value) -> Optional[str]:
    """
    A wrapper over str(), but converts bool values to lower case strings.
    If None is given, just returns None, instead of converting it to string "None".
    """
    if isinstance(value, bool):
        return str(value).lower()
    elif value is None:
        return value
    else:
        return str(value)


class KensuDataFrameWriter:

    def __init__(self,
                 df: "DataFrame",
                 compute_count_distinct: bool = False):
        self._df = df
        self._spark = df.sparkSession
        self._df_writer = DataFrameWriter(df)  # mutable as we modify df at last moment before write
        self._delayed_calls = []
        self._options = {}
        self._format = None
        self._compute_count_distinct = compute_count_distinct

    # FIXME: implement  __getattr__(self, item) for unknown field names to defer to self._df_writer

    def _delay_fn_call(self, fn):
        self._delayed_calls.append(fn)
        return self

    def _update_df(self, new_df):
        self._df = new_df
        self._df_writer = DataFrameWriter(new_df)

    def _call_deferred_fns(self):
        # always returns same self, so it's fine
        for fn in self._delayed_calls:
            fn()

    @staticmethod
    def _path_from_args(args, kwargs):
        """
        Allow to call methods:
        - .parquet("abc.parquet") - args
        - .parquet("abc.parquet", mode="overwrite") - args + kwargs
        - .parquet(path="abc.parquet") - kwargs
        - .parquet(mode="overwrite", path="abc.parquet") - kwargs in weird order - wouldn't work with `def parquet(path, *args, **kwrags)`

        :return: path for first arg or from kwargs
        """
        return kwargs.get('path') or (args and args[0])

    def mode(self, saveMode: Optional[str]) -> "KensuDataFrameWriter":
        self._delay_fn_call(lambda: self._df_writer.mode(saveMode))
        return self

    def format(self, source: str) -> "KensuDataFrameWriter":
        self._format = source
        self._delay_fn_call(lambda: self._df_writer.format(source))
        return self

    def option(self, key: str, value: "OptionalPrimitiveType") -> "KensuDataFrameWriter":
        self._options[key] = to_str(value)
        self._delay_fn_call(lambda: self._df_writer.option(key, to_str(value)))
        return self

    def options(self, **options: "OptionalPrimitiveType") -> "KensuDataFrameWriter":
        for k in options:
            self._options[k] = to_str(options[k])
        self._delay_fn_call(lambda: self._df_writer.options(**options))
        return self

    def _path_from_options(self):
        return self._options.get('path')

    @overload
    def partitionBy(self, *cols: str) -> "KensuDataFrameWriter":
        ...

    @overload
    def partitionBy(self, *cols: List[str]) -> "KensuDataFrameWriter":
        ...

    def partitionBy(self, *cols: Union[str, List[str]]) -> "KensuDataFrameWriter":
        self._delay_fn_call(lambda: self._df_writer.partitionBy(*cols))
        return self

    @overload
    def bucketBy(self, numBuckets: int, col: str, *cols: str) -> "KensuDataFrameWriter":
        ...

    @overload
    def bucketBy(self, numBuckets: int, col: TupleOrListOfString) -> "KensuDataFrameWriter":
        ...

    def bucketBy(
        self, numBuckets: int, col: Union[str, TupleOrListOfString], *cols: Optional[str]
    ) -> "KensuDataFrameWriter":
        self._delay_fn_call(lambda: self._df_writer.bucketBy(numBuckets, col, *cols))
        return self

    @overload
    def sortBy(self, col: str, *cols: str) -> "KensuDataFrameWriter":
        ...

    @overload
    def sortBy(self, col: TupleOrListOfString) -> "KensuDataFrameWriter":
        ...

    def sortBy(
        self, col: Union[str, TupleOrListOfString], *cols: Optional[str]
    ) -> "KensuDataFrameWriter":
        self._delay_fn_call(lambda: self._df_writer.sortBy(col, *cols))
        return self

    def _handle_simple_save(self,
                            path: Optional[str] = None,
                            format: Optional[str] = None,
                            table_name: Optional[str] = None,
                            jdbc_options: Optional[Dict[str, str]] = None
                            ):
        if path is None:
            path = self._path_from_options()
        if path is not None or table_name is not None or jdbc_options is not None:
            df = self._df
            try:
                from kensu.pyspark.spark_connector import addOutputObservationsWithRemoteConf
                import logging
                logging.info("KENSU: DataFrameWriter for output path={} table_name={} format={}, will be automatically updated with Kensu observations via .observe() using remote config if enabled".format(path, table_name, format))
                df = addOutputObservationsWithRemoteConf(df,
                                                         path=path,
                                                         table_name=table_name,
                                                         jdbc_options=jdbc_options,
                                                         format=format,
                                                         compute_count_distinct=self._compute_count_distinct)
                logging.info("KENSU: DataFrameWriter for output path={}  table_name={} format={}, was updated with Kensu observations via .observe() using remote config if enabled".format(path, table_name, format))
            except:
                import traceback
                import logging
                logging.info(
                    "KENSU: unexpected issue when adding output observations to output path={}  table_name={} format={}, are you using old kensu Jar?: {}".format(
                        path, table_name, format,
                        traceback.format_exc()))
            self._update_df(df)
        self._call_deferred_fns()

    def save(
        self,
        path: Optional[str] = None,
        format: Optional[str] = None,
        mode: Optional[str] = None,
        partitionBy: Optional[Union[str, List[str]]] = None,
        **options: "OptionalPrimitiveType",
    ) -> None:
        self._handle_simple_save(path, format or self._format)
        return self._df_writer.save(path=path, format=format, mode=mode, partitionBy=partitionBy, **options)

    def insertInto(self, tableName: str, overwrite: Optional[bool] = None) -> None:
        self._handle_simple_save(path=None, table_name=tableName)
        return self._df_writer.insertInto(tableName, overwrite)

    def saveAsTable(
        self,
        name: str,
        format: Optional[str] = None,
        mode: Optional[str] = None,
        partitionBy: Optional[Union[str, List[str]]] = None,
        **options: "OptionalPrimitiveType",
    ) -> None:
        self._handle_simple_save(path=None, table_name=name, format=format or self._format)
        return self._df_writer.saveAsTable(name=name, format=format, mode=mode, partitionBy=partitionBy, **options)

    def json(self, *args, **kwargs) -> None:
        self._handle_simple_save(path=self._path_from_args(args, kwargs), format="json")
        return self._df_writer.json(*args, **kwargs)

    def parquet(self, *args, **kwargs) -> None:
        self._handle_simple_save(path=self._path_from_args(args, kwargs), format="parquet")
        return self._df_writer.parquet(*args, **kwargs)

    def text(self, *args, **kwargs) -> None:
        self._handle_simple_save(path=self._path_from_args(args, kwargs), format="text")
        return self._df_writer.text(*args, **kwargs)

    def csv(self, *args, **kwargs) -> None:
        self._handle_simple_save(path=self._path_from_args(args, kwargs), format="csv")
        return self._df_writer.csv(*args, **kwargs)

    def orc(self, *args, **kwargs) -> None:
        self._handle_simple_save(path=self._path_from_args(args, kwargs), format="orc")
        return self._df_writer.orc(*args, **kwargs)

    def jdbc(
        self,
        url: str,
        table: str,
        mode: Optional[str] = None,
        properties: Optional[Dict[str, str]] = None,
    ) -> None:
        # Observe is unsupported for jdbc writes due a known Apache Spark bug, see:
        # - https://issues.apache.org/jira/browse/SPARK-42034
        # - https://github.com/apache/spark/pull/39976#issuecomment-1752930380
        # jdbc_options = self._options.copy()
        # jdbc_options.update(properties or {})
        # jdbc_options.update({'url': url, 'dbtable': table})
        # self._handle_simple_save(
        #     jdbc_options=dict([(k, v) for k, v in jdbc_options.items() if isinstance(v, str)]),
        #     format="jdbc")
        return self._df_writer.jdbc(url=url, table=table, mode=mode, properties=properties)



import matplotlib.axes._subplots as sbp
from kensu.utils.wrappers import kensu_wrapper_and_report

class AxesSubplot(sbp.AxesSubplot):
    inheritance = []

    def add_inheritance(self, arg):
        self.inheritance.append(arg)







        # from kensu.utils.wrappers import kensu_wrapper_and_report
        # for f in dir(cls):
        #     if callable(getattr(cls, f)) and f.startswith('__') is False and f != "add_inheritance":
        #         exec("setattr(cls,f, kensu_wrapper_and_report(cls.%s,cls))" % f)



import logging

from kensu.client import DataSourcePK, DataSource, SchemaPK, Schema, FieldDef
from kensu.utils.kensu_provider import KensuProvider
from kensu.utils.helpers import eventually_report_in_mem, flatten, extract_short_json_schema
from kensu.requests.models import Response

import os

import requests as req

def wrap_get(method):
    def wrapper(*args, **kwargs):
        result = method(*args, **kwargs)

        kensu = KensuProvider().instance()

        try:
            # Construct the file location

            url = kwargs['url'] if 'url' in kwargs else args[0]
            try:
                params = kwargs['params'] if 'params' in kwargs else args[1]
            except:
                params = None

            include_url_params = os.environ.get('REQUESTS_INCLUDE_URL_PARAMS', '0') == '1'
            if params is not None and include_url_params:
                location = url + '?' + ('&').join([e+'={}' for e in params.keys()])
            elif not include_url_params:
                # remove url params in case provided hardcoded in url string and not in params
                location = url.split("?")[0]
            else:
                location = url

            result_pk = DataSourcePK(location=location,
                                     physical_location_ref=kensu.default_physical_location_ref)
            result_ds = DataSource(name=location, format='JSON',
                                   pk=result_pk)._report()

            # Construct the real schema
            result_json = result.json()

            fields_dict = flatten(result_json)

            fields = [FieldDef(name = k,  field_type=fields_dict[k], nullable=True ) for k in  fields_dict]
            sc_pk = SchemaPK(result_ds.to_ref(),
                             fields=fields)

            real_sc = Schema(name="schema:" + result_ds.name, pk=sc_pk)

            #Construct the concise schema
            result_sc = extract_short_json_schema(result_json, result_ds)._report()

            if not kensu.lean_mode:
                result.__class__ = Response
                result.ksu_short_schema = result_sc
                result.ksu_schema = real_sc
                result.ds_location = result.url
            try:
                import json
                d = json.loads(result.text)
                count_json = len(d)
                stats = {'count':count_json}

            except:
                stats = None
            kensu.real_schema_df[result_sc.to_guid()] = None
            result.ksu_stats = stats

            if kensu.lean_mode:
                kensu.register_input_lean_mode(result_sc)
        except:
            pass

        return result

    wrapper.__doc__ = method.__doc__
    return wrapper

get = wrap_get(req.get)



# this file should not use f-strings syntax
# because f-strings are not supported in python 3.5
# so instead of: f"my formatted {text}" # py3.6+
# we use: "my formatted {}".format(text) # py3.5+

import logging
from kensu.utils.kensu_conf_file.conf_file import extract_config_property, get_conf_path

# This function takes the fully classified object name, say: <package>.<name>.
# Returns the static object instance on the heap
def ref_scala_object(jvm, object_name):
    reflection = False
    if reflection:
      clazz = jvm.java.lang.Class.forName(object_name + "$")
      ff = clazz.getDeclaredField("MODULE$")
      o = ff.get(None)
      return o
    else:
      # it seems in paid databricks (sometimes) reflection do not work...
      class_name_parts = (object_name + "$.MODULE$").split(".")
      o = jvm
      for p in class_name_parts:
        o = getattr(o, p)
      return o


def wait_for_spark_completion(spark, wait_timeout_secs):
    sc = spark.sparkContext
    jvm = sc._jvm
    client_class = ref_scala_object(jvm, "io.kensu.sparkcollector.KensuSparkCollector")
    duration_class = ref_scala_object(jvm, "scala.concurrent.duration.Duration")
    duration_30s = duration_class.apply(30, "second")
    duration_wait_timeout = duration_class.apply(wait_timeout_secs, "second")
    return client_class.shutdown(duration_30s, duration_wait_timeout)


def patched_spark_stop(wrapped, wait_timeout_secs):
    def wrapper(self):
        try:
            logging.info('KENSU: in spark.stop, waiting for Kensu Spark-Collector reporting to finish')
            wait_for_spark_completion(self, wait_timeout_secs)
            logging.info('KENSU: in spark.stop, waiting for Kensu Spark-Collector done, stopping spark context')
        except:
            import traceback
            logging.warning("KENSU: unexpected issue: {}".format(traceback.format_exc()))
        result = wrapped(self)
        logging.info('KENSU: in spark.stop, spark context stopped')
        return result

    wrapper.__doc__ = wrapped.__doc__
    return wrapper


def rand_str():
    import random
    import string
    return ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(8))


def ensure_dir_exists(path):
    import os
    import errno
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

def patched_spark_createDataFrame_pandas(wrapped, pandas_to_spark_df_via_tmp_file, tmp_dir=None):
    def wrapper(self, data, *args, **kwargs):
        try:
            from kensu import pandas as ksu_pandas
            if isinstance(data, ksu_pandas.data_frame.DataFrame):
                logging.info('KENSU: replacing kensu.pandas.data_frame.DataFrame wrapper into a original pandas...')
                if not pandas_to_spark_df_via_tmp_file:
                    #  just unwrap the Kensu wrapper, so spark at least works
                    data = data.get_df() # type: pandas.DataFrame
                    # to ensure old-style spark.createDataFrame works
                    # even if passed extra kwarg when KSU tmp file is disabled
                    # we delete the key ksu_tmp_file_prefix if it exists
                    kwargs.pop('ksu_tmp_file_prefix', None)
                else:
                    # given `spark_df=spark.createDataFrame(pandas_df)`, it would be replaced into:
                    # res.to_parquet('/tmp/pandas_to_spark/some_tmp_file.parquet', index=False)
                    # spark_df = spark.read.parquet('file:/tmp/pandas_to_spark/some_tmp_file.parquet')
                    tmp_file_prefix = kwargs.get('ksu_tmp_file_prefix') or 'pandasdf_to_spark'
                    tmp_file_name = tmp_file_prefix + '-' + rand_str()
                    import os
                    # create a unique dir with unique file, so both LDS and file are unique - thus do not clash in Kensu
                    final_tmp_dir = os.path.join(tmp_dir, tmp_file_name)
                    ensure_dir_exists(final_tmp_dir)
                    tmp_file_path = os.path.join(final_tmp_dir, tmp_file_name + '.parquet')
                    # write to a temp file locally with pandas, using Kensu wrapper to capture the lineage
                    abs_tmp_file_name = os.path.abspath(tmp_file_path)
                    logging.info('KENSU: spark.createDataFrame(pandas_df): Kensu will write pandas_df to a temp file - '+ abs_tmp_file_name)
                    data.to_parquet(abs_tmp_file_name, index=False)
                    # here do not call the Spark's original implementation of spark.createDataFrame(),
                    # and just read tmp file with spark, so we'd have the lineage
                    spark_temp_file_path = 'file:' + abs_tmp_file_name
                    logging.info('KENSU: spark.createDataFrame(pandas_df): Kensu will read to a temp file in spark - ' + spark_temp_file_path)
                    spark_df = self.read.parquet(spark_temp_file_path)
                    return spark_df
            logging.info('KENSU: done modifying arguments to spark.createDataFrame')
        except:
            import traceback
            logging.warning("KENSU: unexpected issue: {}".format(traceback.format_exc()))
        result = wrapped(self, data, *args, **kwargs)
        logging.info('KENSU: end of spark.createDataFrame')
        return result

    wrapper.__doc__ = wrapped.__doc__
    return wrapper


def patched_dataframe_write(wrapped):
    def wrapper(self):
        try:
            logging.info('KENSU: in patched DataFrame.write, marking DataFrame as cached')
            from pyspark.sql import DataFrameWriter
            result = DataFrameWriter(self.cache())
            logging.info('KENSU: in patched DataFrame.write, returning result')
        except:
            import traceback
            logging.warning("KENSU: unexpected issue in DataFrame.write: {}".format(traceback.format_exc()))
            from pyspark.sql import DataFrameWriter
            result = DataFrameWriter(self.cache())
        return result

    return property(wrapper)


def patched_dataframe_toPandas(wrapped):
    def wrapper(self):
        try:
            logging.info('KENSU: in patched DataFrame.toPandas()')
            result = wrapped(self)
            logging.info('KENSU: in patched DataFrame.toPandas(), returning result')
        except:
            import traceback
            logging.warning("KENSU: unexpected issue in DataFrame.toPandas(): {}".format(traceback.format_exc()))
            result = wrapped(self)
        return result

    return wrapper


def get_spark_from_df(dataframe  # type: DataFrame
                    ):
    sql_context = dataframe.sql_ctx  # type: SQLContext
    return sql_context.sparkSession


def get_jvm_from_df(dataframe  # type: DataFrame
                    ):
    spark = get_spark_from_df(dataframe)
    return spark.sparkContext._jvm


def call_df_as_kensu_datasource(df,  # type: DataFrame
                                location, datasource_type,
                                name=None,  # type: str
                                logical_name=None  # type: str
                                ):
    jvm = get_jvm_from_df(df)
    w = jvm.io.kensu.sparkcollector.KensuSparkCollector.KensuSparkDataFrame(df._jdf)
    return w.reportAsKensuDatasource(location, datasource_type, jvm.scala.Option.apply(name), jvm.scala.Option.apply(logical_name))


def call_df_as_kensu_jdbc_datasource(df,  # type: DataFrame
                                     dbType,  # type: str
                                     schemaName,  # type: str
                                     tableName,  # type: str
                                     maybeDatabaseName,  # type: str
                                     name=None,  # type: str
                                     logical_name=None  # type: str
                                     ):
    jvm = get_jvm_from_df(df)
    w = jvm.io.kensu.sparkcollector.KensuSparkCollector.KensuSparkDataFrame(df._jdf)
    return w.reportAsKensuJdbcDatasource(dbType, schemaName, tableName, jvm.scala.Option.apply(maybeDatabaseName), jvm.scala.Option.apply(name), jvm.scala.Option.apply(logical_name))


def do_disable_spark_writes(spark):
    """
    Disables spark writes, but keeps Kensu Spark-Collector reporting.
    Affects only the writes which use DataFrameWriter - dataFrame.write.someMethod(), including .csv(), .save(), .saveAsTable().

    Note: This function MUST be run before first spark write, otherwise spark writes will not be disabled!
    """
    logging.info('KENSU: disabling df.write')
    sc = spark.sparkContext
    jvm = sc._jvm
    spark_writes_disabler = ref_scala_object(jvm, "io.kensu.sparkcollector.spark.nowrites.SparkWritesDisabler")
    spark_writes_disabler.disableSparkWrites()
    logging.info('KENSU: disabling df.write done')


def set_fake_timestamp(spark, mocked_timestamp):
    """
    Sets a fake timestamp to be reported to Kensu Spark-Collector
    """
    logging.info('KENSU: overriding execution_timestamp starting')
    sc = spark.sparkContext
    jvm = sc._jvm
    clazz = ref_scala_object(jvm, "io.kensu.sparkcollector.utils.TimeUtils")
    force_override = False
    clazz.setMockedTime(int(mocked_timestamp), force_override)
    logging.info('KENSU: overriding execution_timestamp done')


def add_ds_path_sanitizer(spark, search_str, replacement_str):
    logging.info('KENSU: add_ds_path_sanitizer starting')
    sc = spark.sparkContext
    jvm = sc._jvm
    clazz = ref_scala_object(jvm, "io.kensu.sparkcollector.datasources.DataSourceConverters")
    clazz.addDsPathSanitizer(search_str, replacement_str)
    logging.info('KENSU: add_ds_path_sanitizer done')

def report_df_as_kensu_datasource():
    def report_as_kensu_datasource(self, # type: DataFrame
                                 location, # type: str
                                 datasource_type=None, # type: str
                                 name=None, # type: str
                                 logical_name=None # type: str
                                 ):
        try:
            logging.info('KENSU: in df.report_as_kensu_datasource')
            call_df_as_kensu_datasource(self, location, datasource_type or 'in-memory-dataframe', name, logical_name)
            logging.info('KENSU: df.report_as_kensu_datasource started in background')
        except:
            import traceback
            logging.warning("KENSU: unexpected issue: {}".format(traceback.format_exc()))
        return self
    return report_as_kensu_datasource


def report_df_as_kensu_jdbc_datasource():
    def report_as_kensu_jdbc_datasource(self,  # type: DataFrame
                                      db_type,  # type: str
                                      schema_name,  # type: str
                                      table_name,  # type: str
                                      maybe_db_name=None,  # type: str
                                      name=None, # type: str
                                      logical_name=None # type: str
                                      ):
        try:
            logging.info('KENSU: in df.report_as_kensu_jdbc_datasource')
            call_df_as_kensu_jdbc_datasource(self, db_type, schema_name, table_name, maybe_db_name, name, logical_name)
            logging.info('KENSU: df.report_as_kensu_jdbc_datasource started in background')
        except:
            import traceback
            logging.warning("KENSU: unexpected issue: {}".format(traceback.format_exc()))
        return self
    return report_as_kensu_jdbc_datasource


def report_df_as_kpi():
    def report_as_kpi(self,  # type: DataFrame
                      name,  # type: str
                      logical_name=None,  # type: str
                      datasource_type="Business metric (KPI)" # type: str
                      ):
        try:
            logging.info('KENSU: in df.report_as_kpi')
            jvm = get_jvm_from_df(self)
            w = jvm.io.kensu.sparkcollector.KensuSparkCollector.KensuSparkDataFrame(self._jdf)
            w.reportAsKPI(name, logical_name or name, datasource_type)
            logging.info('KENSU: df.report_as_kpi started in background')
        except:
            import traceback
            logging.warning("KENSU: unexpected issue: {}".format(traceback.format_exc()))
        return self
    return report_as_kpi


def patch_kensu_df_helpers(patch_spark_data_frame_write,
                           kensu_efficient_write_compute_count_distinct):
    kensu_efficient_write_fn = make_kensu_efficient_write_fn(
        kensu_efficient_write_compute_count_distinct=kensu_efficient_write_compute_count_distinct)
    optional_patches = []
    if patch_spark_data_frame_write:
        optional_patches.append(['write', kensu_efficient_write_fn])
    fns_to_patch = list(map(lambda fn: [fn.__name__, fn], [
        addOutputObservations,
        addCustomObservationsToOutput,
        reportCustomKafkaInputSchema,
        reportCustomEventhubInputSchema
    ])) + [
        ['report_as_kensu_datasource', report_df_as_kensu_datasource()],
        ['report_as_kensu_jdbc_datasource', report_df_as_kensu_jdbc_datasource()],
        ['report_as_kpi', report_df_as_kpi()],
        ['kensu_efficient_write', kensu_efficient_write_fn]
    ] + optional_patches
    for fn_name, wrapper_builder in fns_to_patch:
        try:
            logging.info('KENSU: Patching DataFrame.{}'.format(fn_name))
            from pyspark.sql import DataFrame
            # can setattr behave differently than .somename = somevalue?
            setattr(DataFrame, fn_name, wrapper_builder)
            logging.info('KENSU: done patching DataFrame.{}'.format(fn_name))
        except:
            import traceback
            log_and_print(logging.warning,
                          "KENSU: unexpected issue when patching DataFrame.{}: {}".format(
                              fn_name, traceback.format_exc()
                          ))


def join_paths(maybe_directory, # type: str
               file # type: str
               ):
    if maybe_directory is not None and maybe_directory.strip():
        import os
        return os.path.join(maybe_directory, file)
    else:
        return file


def convert_naming_rules(rules # type: list[str]
                         ):
    # a rule looks like this: kensu-spark-collector/hive-query-results->>File
    rules_list = list(["{}->>{}".format(matcher, formatter) for (matcher, formatter) in rules])
    # FIXME: allow passing a list in spark-collector
    return ";".join(rules_list)


def scala_ds_and_schema_to_py(kensu_instance, ds, schema):
    from kensu.utils.dsl.extractors.external_lineage_dtos import KensuDatasourceAndSchema
    from kensu.client import DataSourcePK, DataSource, FieldDef, SchemaPK, Schema
    pl_ref = kensu_instance.UNKNOWN_PHYSICAL_LOCATION.to_ref()
    py_ds = DataSource(name=ds.name(),
                       format=ds.format() or None,
                       categories=list(ds.categories() or []), # NULLABLE
                       pk=DataSourcePK(location=ds.pk().location(), physical_location_ref=pl_ref))

    # schema
    fields = [
        FieldDef(name=f.name(), field_type=f.fieldType(), nullable=f.nullable())
        for f in schema.fields()
    ]
    py_schema = Schema(
        name=schema.name(),
        pk=SchemaPK(py_ds.to_ref(),
                    fields=fields))

    def f_publish_stats(lineage_run_id):
        schema.attachStatsTo(lineage_run_id)

    def f_get_stats():
        # ******************************************************************************************
        # WARNING : Do not return anything else than None, because f_publish_stats is preferred giving the responsibility
        # to Spark to send the data stats only when it's available (when computation inside scala Future is completed)
        # ******************************************************************************************
        return None
    return KensuDatasourceAndSchema(ksu_ds=py_ds, ksu_schema=py_schema, f_get_stats=f_get_stats, f_publish_stats = f_publish_stats)


def j2py_dict_of_lists(jdict):
    r = {}
    for k, jlist in jdict.items():
        r[k] = list(jlist)
    return r


def get_inputs_lineage_fn(kensu_instance, df, virtual_ds_name=None, virtual_ds_logical_name=None):
    jvm = get_jvm_from_df(df)
    # call def fetchToPandasReport(
    #     df: DataFrame,
    #     name: String = UUID.randomUUID().toString,
    #     logicalName: Option[String] = None,
    #     datasourceType: String = InternalDatasourceTags.TAG_INPUT_LINEAGE_ONLY,
    #     timeout: Duration
    #   )
    client_class = ref_scala_object(jvm, "io.kensu.sparkcollector.dataconversions.SparkToPandasConversionTracker")
    duration_300s =  ref_scala_object(jvm, "scala.concurrent.duration.Duration").apply(300, "second")
    # FIXME: it should NOT wait for datasts -> check
    import uuid
    if not virtual_ds_name:
        virtual_ds_name = str(uuid.uuid1())
    # FIXME: virtual_ds_logical_name
    # FIXME: add format
    no_logical_name = jvm.scala.Option.apply(None)
    lineage_from_scala = client_class.fetchToPandasReport(
        df._jdf,
        virtual_ds_name,
        no_logical_name,
        # ds_type,
        duration_300s
    )
    # logging.info('lineage_from_scala:', lineage_from_scala)
    df.lineage_from_scala = lineage_from_scala
    from kensu.utils.dsl.extractors.external_lineage_dtos import GenericComputedInMemDs, ExtDependencyEntry
    # case class SimpleInputDsSchemaAndLineage(
    #   ds: DataSource,
    #   schema: Schema,
    #   lineage: Map[String, Seq[String]]
    # )
    #
    # case class SimpleInputsLineageReport(
    #   lineage: Seq[SimpleInputDsSchemaAndLineage]
    # )
    lineage_info = [
        ExtDependencyEntry(
            input_ds=scala_ds_and_schema_to_py(kensu_instance, e.ds(), e.schema()),
            lineage=j2py_dict_of_lists(e.lineage())
        ) for e in lineage_from_scala.lineage()
    ]
    logging.info('KENSU: lineage_info:', lineage_info)
    return GenericComputedInMemDs(inputs=list([x.input_ds for x in lineage_info]), lineage=lineage_info)


def get_process_run_info(spark):
    if spark is not None:
        logging.info('Getting Process info from Kensu Spark Agent')
        cls = ref_scala_object(spark.sparkContext._jvm, "io.kensu.sparkcollector.pyspark.PySparkUtils")
        # requires Kensu collector to be already initialized - if not yet, it returns None
        # getProcessAndRunGuid(spark): ProcessRunInfo (nullable)
        # case class ProcessRunInfo(processGuid: String, processRunGuid: String)
        run_info = cls.getProcessAndRunGuid(spark._jsparkSession)
        if run_info:
            logging.info('Successfully retrieved the Process info from Kensu Spark Agent')
            return {
                "process_guid": run_info.processGuid(),
                "process_run_guid": run_info.processRunGuid()
            }
    logging.info('Unable to get Process info from Kensu Spark Agent, maybe Spark Agent is not initialized yet?')
    return None


def get_df_with_inmem_tag(df,  # type: DataFrame
                          df_tag  # type: str
                          ):
    jvm = get_jvm_from_df(df)
    cls = ref_scala_object(jvm, "io.kensu.sparkcollector.lineage.CreateDataFrameLineage")
    # def dataframeWithLineageTag(df: DataFrame, inputsId: String): DataFrame
    jdf = cls.dataframeWithLineageTag(df._jdf, df_tag)
    # finally convert Java DataFrame back to python DataFrame
    spark = get_spark_from_df(df)
    from pyspark.sql.dataframe import DataFrame
    return DataFrame(jdf, spark)


def register_manual_lineage(spark,
                            output_df_tag,  # type: str
                            inputs          # type: list[dict[str, object]]
                            ):
    # def registerLineage(spark: SparkSession, inputsId: String, inputDatasources: java.util.List[SimpleInputDsSchema])
    cls_reporter = ref_scala_object(spark.sparkContext._jvm, "io.kensu.sparkcollector.lineage.CreateDataFrameLineage")
    # build inputDatasources
    # case class SimpleInputDsSchema(path: String, format: String, schema: java.util.Map[String, String])
    cls_inputobj = ref_scala_object(spark.sparkContext._jvm, "io.kensu.sparkcollector.lineage.SimpleInputDsSchema")
    inputDatasources = list([
        cls_inputobj.apply(i.get('path'), i.get('format'), i.get('schema'))
        for i in inputs
    ])
    cls_reporter.registerLineage(spark._jsparkSession, output_df_tag, inputDatasources)

"""
Assuming Datasource name is /a/b/c.ext
Allowed values for parameters:
- data_source_naming_strategy:
  * None (default) - (file) name of DS if use_short_datasource_names else full_path
  * 'File' - c.ext
  * 'LastTwoFoldersAndFile' - a/b/c.ext
  * 'LastFolderAndFile' - b/c.ext
  * 'LastFolder' - b
  * 'PathBasedRule' - use data_source_naming_strategy_rules
- logical_data_source_naming_strategy:
  * None (default) - do not set logical datasource
  * 'File' - c.ext
  * 'LastTwoFoldersAndFile' - a/b/c.ext
  * 'LastFolderAndFile' - b/c.ext
  * 'LastFolder' - b
  * 'PathBasedRule' - use logical_datasource_name_strategy_rules
- data_source_naming_strategy_rules or logical_data_source_naming_strategy_rules:
  * pass a list of tuples (string_matcher, naming_strategy) like this:
    [("features_with_attrition_horizon_label", "File"),
     ("attri_pro_risk_indicator_disabled_3", "LastTwoFoldersAndFile"),
     # default - empty string matches everything
     ("", "LastFolderAndFile")]
- missing_column_lineage_strategy:
    * 'CaseInsensitiveColumnNameMatcherStrat'
    * 'CaseSensitiveColumnNameMatcherStrat'
    * 'AllToAllLineageStrat'
    * 'OutFieldEndsWithInFieldNameLineageStrat'
    * 'OutFieldStartsWithInFieldNameLineageStrat'
"""
def init_kensu_spark(
        spark_session=None,  # Must stay first as mostly this is the only argument passed
        kensu_ingestion_url=None,
        kensu_ingestion_token=None,
        report_to_file=None,
        disable_tracking=False,
        logs_dir_path=None,
        offline_file_name=None,
        process_name=None,
        process_run_name=None,
        code_location=None,
        code_version=None,
        user_name=None,
        enable_entity_compaction=None,
        kensu_api_verify_ssl=None,
        compute_stats=None,
        compute_input_stats=None,
        compute_output_stats=None,
        input_stats_for_only_used_columns=None,
        input_stats_keep_filters=None,
        input_stats_compute_quantiles=None,
        input_stats_compute_simple_num=None,
        input_stats_compute_std_dev=None,
        input_stats_cache_by_path=None,
        input_stats_coalesce_enabled=None,
        input_stats_coalesce_workers=None,
        output_stats_compute_quantiles=None,
        output_stats_compute_simple_num=None,
        output_stats_compute_std_dev=None,
        output_stats_cache_by_path=None,
        output_stats_coalesce_enabled=None,
        output_stats_coalesce_workers=None,
        cache_output_for_stats=None,
        shorten_data_source_names=None,
        shutdown_timeout_sec=None,
        enable_collector_log_file=None,
        collector_log_level=None,
        collector_log_include_spark_logs=None,
        data_source_naming_strategy=None,
        data_source_naming_strategy_rules=None,
        logical_data_source_naming_strategy=None,
        logical_data_source_naming_strategy_rules=None,
        missing_column_lineage_strategy=None,
        patch_spark_data_frame=None,
        patch_spark_data_frame_write=None,
        kensu_efficient_write_compute_count_distinct=None,
        disable_spark_writes=None,
        environment=None,
        execution_timestamp=None,
        project_name=None,
        h2o_support=None,
        h2o_create_virtual_training_datasource=None,
        patch_pandas_conversions=False,
        pandas_to_spark_df_via_tmp_file=None,
        pandas_to_spark_tmp_dir=None,
        use_api_client=None,
        remote_conf_enabled=True,
        remote_conf_timeout_secs=None,
        remote_circuit_breaker_enabled=True,
        remote_circuit_breaker_precheck_delay_secs=None,
        datastats_send_timeout_secs=None,
        conf_file_path=None,
        **kwargs
):
    import os
    is_tracking_disabled = (os.environ.get("KSU_DISABLE_TRACKING", 'False') == 'True') or disable_tracking

    if not is_tracking_disabled:
        logging.info("Initializing Kensu tracking...")
        jvm = spark_session.sparkContext._jvm

        from configparser import ConfigParser, ExtendedInterpolation
        config = ConfigParser(interpolation=ExtendedInterpolation())
        conf_path = get_conf_path("conf.ini")
        try:
            config.read(conf_path)
        except:
            logging.warning("Cannot load config from file `%s`" % (conf_path))

        kensu_conf = config['kensu'] if config.has_section('kensu') else config['DEFAULT']

        use_api_client = extract_config_property('use_api_client', False, use_api_client, kw=kwargs, conf=kensu_conf, tpe=bool)

        kensu_ingestion_url = extract_config_property('kensu_ingestion_url', None, kensu_ingestion_url, kw=kwargs, conf=kensu_conf)
        kensu_ingestion_token = extract_config_property('kensu_ingestion_token', None, kensu_ingestion_token, kw=kwargs, conf=kensu_conf)
        report_to_file = extract_config_property('report_to_file', False, report_to_file, kw=kwargs, conf=kensu_conf, tpe=bool)
        logs_dir_path = extract_config_property('logs_dir_path', None, logs_dir_path, kw=kwargs, conf=kensu_conf)
        offline_file_name = extract_config_property('offline_file_name', 'kensu-offline.log', offline_file_name, kw=kwargs, conf=kensu_conf)

        shutdown_timeout_sec = extract_config_property('shutdown_timeout_sec', 10 * 60, shutdown_timeout_sec, kw=kwargs, conf=kensu_conf, tpe=int)
        kensu_api_verify_ssl = extract_config_property('kensu_api_verify_ssl', True, kensu_api_verify_ssl, kw=kwargs, conf=kensu_conf, tpe=bool)
        enable_collector_log_file = extract_config_property('enable_collector_log_file', False, enable_collector_log_file, kw=kwargs, conf=kensu_conf, tpe=bool)
        collector_log_level = extract_config_property('collector_log_level', 'INFO', collector_log_level, kw=kwargs, conf=kensu_conf)
        collector_log_include_spark_logs = extract_config_property('collector_log_include_spark_logs', False, collector_log_include_spark_logs, kw=kwargs, conf=kensu_conf, tpe=bool)

        process_name = extract_config_property('process_name', None, process_name, kw=kwargs, conf=kensu_conf)
        process_run_name = extract_config_property('process_run_name', None, process_run_name, kw=kwargs, conf=kensu_conf)
        code_location = extract_config_property('code_location', None, code_location, kw=kwargs, conf=kensu_conf)
        code_version = extract_config_property('code_version', None, code_version, kw=kwargs, conf=kensu_conf)
        user_name = extract_config_property('user_name', None, user_name, kw=kwargs, conf=kensu_conf)
        enable_entity_compaction = extract_config_property('enable_entity_compaction',True,enable_entity_compaction, kw=kwargs, conf=kensu_conf, tpe=bool)
        environment = extract_config_property('environment',None,environment, kw=kwargs, conf=kensu_conf)
        project_name = extract_config_property('project_name', None, project_name, kw=kwargs, conf=kensu_conf)
        execution_timestamp = extract_config_property('execution_timestamp', None, execution_timestamp, kw=kwargs, conf=kensu_conf, tpe=int)

        compute_stats = extract_config_property('compute_stats', True, compute_stats, kw=kwargs, conf=kensu_conf, tpe=bool)
        compute_input_stats = extract_config_property('compute_input_stats', True, compute_input_stats, kw=kwargs, conf=kensu_conf, tpe=bool)
        compute_output_stats = extract_config_property('compute_output_stats', True, compute_output_stats, kw=kwargs, conf=kensu_conf, tpe=bool)
        input_stats_for_only_used_columns = extract_config_property('input_stats_for_only_used_columns',True,input_stats_for_only_used_columns, kw=kwargs, conf=kensu_conf, tpe=bool)
        input_stats_keep_filters = extract_config_property('input_stats_keep_filters',True,input_stats_keep_filters, kw=kwargs, conf=kensu_conf, tpe=bool)
        input_stats_compute_quantiles = extract_config_property('input_stats_compute_quantiles',False,input_stats_compute_quantiles, kw=kwargs, conf=kensu_conf, tpe=bool)
        input_stats_compute_simple_num = extract_config_property('input_stats_compute_simple_num',None,input_stats_compute_simple_num, kw=kwargs, conf=kensu_conf, tpe=bool)
        input_stats_compute_std_dev = extract_config_property('input_stats_compute_std_dev',None,input_stats_compute_std_dev, kw=kwargs, conf=kensu_conf, tpe=bool)
        input_stats_cache_by_path = extract_config_property('input_stats_cache_by_path', True, input_stats_cache_by_path, kw=kwargs, conf=kensu_conf, tpe=bool)
        input_stats_coalesce_enabled = extract_config_property('input_stats_coalesce_enabled',True,input_stats_coalesce_enabled, kw=kwargs, conf=kensu_conf, tpe=bool)
        input_stats_coalesce_workers = extract_config_property('input_stats_coalesce_workers',1,input_stats_coalesce_workers, kw=kwargs, conf=kensu_conf, tpe=int)
        output_stats_compute_quantiles = extract_config_property('output_stats_compute_quantiles',False,output_stats_compute_quantiles, kw=kwargs, conf=kensu_conf, tpe=bool)
        output_stats_compute_simple_num = extract_config_property('output_stats_compute_simple_num',None,output_stats_compute_simple_num, kw=kwargs, conf=kensu_conf, tpe=bool)
        output_stats_compute_std_dev = extract_config_property('output_stats_compute_std_dev',None,output_stats_compute_std_dev, kw=kwargs, conf=kensu_conf, tpe=bool)
        output_stats_cache_by_path = extract_config_property('output_stats_cache_by_path', False, output_stats_cache_by_path, kw=kwargs, conf=kensu_conf, tpe=bool)
        output_stats_coalesce_enabled = extract_config_property('output_stats_coalesce_enabled',False,output_stats_coalesce_enabled, kw=kwargs, conf=kensu_conf, tpe=bool)
        output_stats_coalesce_workers = extract_config_property('output_stats_coalesce_workers',100,output_stats_coalesce_workers, kw=kwargs, conf=kensu_conf, tpe=int)
        cache_output_for_stats = extract_config_property('cache_output_for_stats', False, cache_output_for_stats, kw=kwargs, conf=kensu_conf, tpe=bool)

        shorten_data_source_names = extract_config_property('shorten_data_source_names', True, shorten_data_source_names, kw=kwargs, conf=kensu_conf, tpe=bool)
        data_source_naming_strategy = extract_config_property('data_source_naming_strategy', None, data_source_naming_strategy, kw=kwargs, conf=kensu_conf)
        data_source_naming_strategy_rules = extract_config_property('data_source_naming_strategy_rules', None, data_source_naming_strategy_rules, kw=kwargs, conf=kensu_conf)
        logical_data_source_naming_strategy = extract_config_property('logical_data_source_naming_strategy', None, logical_data_source_naming_strategy, kw=kwargs, conf=kensu_conf)
        logical_data_source_naming_strategy_rules = extract_config_property('logical_data_source_naming_strategy_rules', None, logical_data_source_naming_strategy_rules, kw=kwargs, conf=kensu_conf)

        missing_column_lineage_strategy = extract_config_property('missing_column_lineage_strategy', None, missing_column_lineage_strategy, kw=kwargs, conf=kensu_conf)

        disable_spark_writes = extract_config_property('disable_spark_writes',False,disable_spark_writes, kw=kwargs, conf=kensu_conf, tpe=bool)

        h2o_support = extract_config_property('h2o_support', False, h2o_support, kw=kwargs, conf=kensu_conf, tpe=bool)
        h2o_create_virtual_training_datasource = extract_config_property('h2o_create_virtual_training_datasource',True,h2o_create_virtual_training_datasource, kw=kwargs, conf=kensu_conf, tpe=bool)

        patch_spark_data_frame = extract_config_property('patch_spark_data_frame', True, patch_spark_data_frame, kw=kwargs, conf=kensu_conf, tpe=bool)
        patch_spark_data_frame_write = extract_config_property('patch_spark_data_frame_write', True, patch_spark_data_frame_write, kw=kwargs, conf=kensu_conf, tpe=bool)
        kensu_efficient_write_compute_count_distinct = extract_config_property('kensu_efficient_write_compute_count_distinct', False, kensu_efficient_write_compute_count_distinct, kw=kwargs, conf=kensu_conf, tpe=bool)

        patch_pandas_conversions = extract_config_property('patch_pandas_conversions', False, patch_pandas_conversions, kw=kwargs, conf=kensu_conf, tpe=bool)
        pandas_to_spark_df_via_tmp_file = extract_config_property('pandas_to_spark_df_via_tmp_file',True,pandas_to_spark_df_via_tmp_file, kw=kwargs, conf=kensu_conf, tpe=bool)
        pandas_to_spark_tmp_dir = extract_config_property('pandas_to_spark_tmp_dir','/tmp/spark-to-pandas-tmp',pandas_to_spark_tmp_dir, kw=kwargs, conf=kensu_conf)

        try:
            ###  Get notebook name ...
            #### see https://github.com/jupyter/notebook/issues/1000#issuecomment-359875246
            import json
            import os.path
            import re
            import requests
            try:  # Python 3
                from urllib.parse import urljoin
            except ImportError:  # Python 2
                from urlparse import urljoin

            def get_notebook_name():
                """
                Return the full path of the jupyter notebook.
                """
                try:  # Python 3
                    from notebook.notebookapp import list_running_servers
                except ImportError:  # Python 2
                    try:
                        import warnings
                        from IPython.utils.shimmodule import ShimWarning
                        with warnings.catch_warnings():
                            warnings.simplefilter("ignore", category=ShimWarning)
                            from IPython.html.notebookapp import list_running_servers
                    except ImportError:  # Probably pyspark script is run without IPython/Jupyter
                        if not process_name:
                            logging.warning('WARN Unable to automatically extract Jupyter/pyspark notebook name (did you run it without jupyter?)')
                        return ['', process_name or 'Unknown pyspark filename']
                try:
                    import ipykernel
                    kernel_id = re.search('kernel-(.*).json',
                                          ipykernel.connect.get_connection_file()).group(1)
                    servers = list_running_servers()
                    for ss in servers:
                        response = requests.get(urljoin(ss['url'], 'api/sessions'),
                                                params={'token': ss.get('token', '')})
                        for nn in json.loads(response.text):
                            if nn['kernel']['id'] == kernel_id:
                                server = ss
                                notebooks_path = server['notebook_dir']
                                return [notebooks_path, nn['notebook']['path']]
                except Exception as e:
                    if not process_name:
                        logging.warning('WARN Unable to automatically extract pyspark notebook name')
                    return ['', process_name or 'Unknown pyspark filename']

            notebooks_path, notebook_name = get_notebook_name()
            logging.info("Notebook name " + notebook_name)

            ### Configuration for tracker
            ingestion_url = jvm.scala.Option.apply(kensu_ingestion_url)
            t2 = jvm.scala.Tuple2
            provider_class_name = t2("spark_environment_provider", "io.kensu.sparkcollector.system.DefaultSparkEnvironmentProvider")

            ### How to get GIT info here?
            #         explicit_code_repo=None,
            #         explicit_code_version=None,
            #         explicit_code_maintainers=None,
            #         explicit_lauched_by_user=None,
            code_repo_name = code_location or "Jupyter-Notebook:spark-" + spark_session.version + "::" + notebooks_path
            repository = t2("code_location", code_repo_name)
            from datetime import datetime
            d = datetime.now()

            code_version_value = code_version or notebook_name + "::" + str(d)
            version = t2("code_version", code_version_value)

            user = t2("user_name", user_name)

            properties = jvm.scala.collection.mutable.HashSet()

            def add_prop(name, value):
                if value is not None:
                    properties.add(t2(name, value))

            properties.add(provider_class_name)
            properties.add(repository)
            properties.add(version)
            properties.add(user)
            if kensu_ingestion_url:
                properties.add(t2("kensu_ingestion_url", kensu_ingestion_url))
            if kensu_ingestion_token:
                properties.add(t2("kensu_ingestion_token", kensu_ingestion_token))
            if report_to_file is not None:
                properties.add(t2("report_to_file", report_to_file))
            if logs_dir_path:
                properties.add(t2("offline_report_dir_path", logs_dir_path))
            if offline_file_name:
                properties.add(t2("offline_file_name", offline_file_name))
            if kensu_api_verify_ssl is not None:
                properties.add(t2("kensu_api_verify_ssl", kensu_api_verify_ssl))
            if enable_entity_compaction is not None:
                properties.add(t2("enable_entity_compaction", enable_entity_compaction))

            if compute_stats is not None:
                properties.add(t2("compute_stats", compute_stats))
            if compute_input_stats is not None:
                properties.add(t2("compute_input_stats", compute_input_stats))
            if compute_output_stats is not None:
                properties.add(t2("compute_output_stats", compute_output_stats))
            if input_stats_for_only_used_columns:
                properties.add(t2("input_stats_for_only_used_columns", input_stats_for_only_used_columns))
            if input_stats_keep_filters is not None:
                properties.add(t2("input_stats_keep_filters", input_stats_keep_filters))

            add_prop("input_stats_compute_quantiles", input_stats_compute_quantiles)
            add_prop("input_stats_compute_simple_num", input_stats_compute_simple_num)
            add_prop("input_stats_compute_std_dev", input_stats_compute_std_dev)
            add_prop("input_stats_cache_by_path", input_stats_cache_by_path)
            add_prop("input_stats_coalesce_enabled", input_stats_coalesce_enabled)
            add_prop("input_stats_coalesce_workers", input_stats_coalesce_workers)

            add_prop("output_stats_compute_quantiles", output_stats_compute_quantiles)
            add_prop("output_stats_compute_simple_num", output_stats_compute_simple_num)
            add_prop("output_stats_compute_std_dev", output_stats_compute_std_dev)
            add_prop("output_stats_cache_by_path", output_stats_cache_by_path)
            add_prop("output_stats_coalesce_enabled", output_stats_coalesce_enabled)
            add_prop("output_stats_coalesce_workers", output_stats_coalesce_workers)

            def extract_prop(name, tpe, value):
                return name, extract_config_property(key=name, default=None, arg=value, kw=kwargs, conf=kensu_conf, tpe=tpe)

            properties_by_name = [
                extract_prop('remote_conf_enabled', tpe=bool, value=remote_conf_enabled),
                extract_prop('remote_conf_timeout_secs', tpe=int, value=remote_conf_timeout_secs),
                extract_prop('remote_circuit_breaker_enabled', tpe=bool, value=remote_circuit_breaker_enabled),
                extract_prop('remote_circuit_breaker_precheck_delay_secs', tpe=int, value=remote_circuit_breaker_precheck_delay_secs),
                extract_prop('datastats_send_timeout_secs', tpe=int, value=datastats_send_timeout_secs),
            ]
            for prop_name, prop_value in properties_by_name:
                if prop_value is not None:
                    add_prop(prop_name, prop_value)

            if shutdown_timeout_sec is not None:
                properties.add(t2("shutdown_timeout_sec", shutdown_timeout_sec))
            if enable_collector_log_file:
                debug_level = collector_log_level or "INFO"
                if process_name is not None:
                    kensu_debug_filename = join_paths(logs_dir_path, process_name + ".kensu-collector.log")
                else:
                    notebook_file_name = notebook_name.split('/')[-1].split('\\')[-1]  # remove path from notebook name
                    kensu_debug_filename = join_paths(logs_dir_path, notebook_file_name + ".kensu-collector.log")
                properties.add(t2("collector_log_level", debug_level))
                properties.add(t2("collector_log_file_path", kensu_debug_filename))
                properties.add(t2("collector_log_include_spark_logs", collector_log_include_spark_logs))
                logging.info("KENSU: Will write Kensu Spark Collector logs to file: " + kensu_debug_filename)
            if shorten_data_source_names is not None:
                properties.add(t2("shorten_data_source_names", shorten_data_source_names))

            if data_source_naming_strategy_rules is not None:
                data_source_naming_strategy = 'PathBasedRule'
                converted_rules = convert_naming_rules(data_source_naming_strategy_rules)
                properties.add(t2("data_source_naming_strategy_rules", converted_rules))
            if logical_data_source_naming_strategy_rules is not None:
                logical_data_source_naming_strategy = 'PathBasedRule'
                converted_rules = convert_naming_rules(logical_data_source_naming_strategy_rules)
                properties.add(t2("logical_data_source_naming_strategy_rules", converted_rules))

            if missing_column_lineage_strategy is not None:
                properties.add(t2('missing_column_lineage_strategy', missing_column_lineage_strategy))

            if data_source_naming_strategy is not None:
                properties.add(t2("data_source_naming_strategy", data_source_naming_strategy))
            if logical_data_source_naming_strategy is not None:
                properties.add(t2("logical_data_source_naming_strategy", logical_data_source_naming_strategy))
            if environment is not None:
                properties.add(t2("environment", environment))
            if project_name is not None:
                properties.add(t2("project_name", project_name))

            process_name = None
            if process_name is not None:
                process_name = process_name
            elif notebook_name != 'Unknown pyspark filename':
                process_name = notebook_name
            if process_name is not None:
                properties.add(t2("process_name", process_name))

            if process_run_name is not None:
                properties.add(t2("process_run_name", process_run_name))

            maybe_fake_timestamp = execution_timestamp
            if maybe_fake_timestamp is not None:
                set_fake_timestamp(spark_session, maybe_fake_timestamp)
            maybe_ds_path_sanitizer_search = os.environ.get('KSU_DS_PATH_SANITIZER_SEARCH')
            maybe_ds_path_sanitizer_replace = os.environ.get('KSU_DS_PATH_SANITIZER_REPLACE')
            if (maybe_ds_path_sanitizer_search is not None) and (maybe_ds_path_sanitizer_replace is not None):
                add_ds_path_sanitizer(spark_session, maybe_ds_path_sanitizer_search, maybe_ds_path_sanitizer_replace)

            if is_databricks():
                print('Detected Databricks Notebook - initializing Kensu via DatabricksCollector.track()')
                w = ref_scala_object(jvm, 'io.kensu.sparkcollector.environments.DatabricksCollector')
                # FIXME: DON'T WE OVERRIDE SOME UNNEEDED CONF KEYS!?
                w.track(spark_session._jsparkSession, conf_file_path or get_conf_path(), properties.toSeq())
            else:
                w = jvm.io.kensu.sparkcollector.KensuSparkCollector.KensuSparkSession(spark_session._jsparkSession)
                w.track(ingestion_url, jvm.scala.Option.empty(), properties.toSeq())

            if (shutdown_timeout_sec is not None) and (shutdown_timeout_sec > 0):
                logging.info('KENSU: patching spark.stop to wait for Kensu Spark-Collector reporting to finish')
                from pyspark.sql import SparkSession
                SparkSession.stop = patched_spark_stop(SparkSession.stop, shutdown_timeout_sec)
                logging.info('KENSU: patching spark.stop done')

            if disable_spark_writes:
                try:
                    do_disable_spark_writes(spark_session)
                except:
                    import traceback
                    logging.warning("KENSU: unexpected issue when disabling df.write {}".format(traceback.format_exc()))

            if cache_output_for_stats and compute_stats:
                try:
                    logging.info('KENSU: patching DataFrame.write to cache/persist the results to speedup data-stats computation')
                    from pyspark.sql import DataFrame
                    DataFrame.write = patched_dataframe_write(DataFrame.write)
                    logging.info('KENSU: done patching DataFrame.write')
                except:
                    import traceback
                    logging.warning("KENSU: unexpected issue when patching DataFrame.write: {}".format(traceback.format_exc()))

            if patch_spark_data_frame:
                patch_kensu_df_helpers(patch_spark_data_frame_write=patch_spark_data_frame_write,
                                       kensu_efficient_write_compute_count_distinct=kensu_efficient_write_compute_count_distinct)

            if patch_pandas_conversions:
                try:
                    logging.info('KENSU: patching DataFrame.toPandas')
                    from pyspark.sql import DataFrame
                    import kensu.pandas as pd
                    DataFrame.toPandas = pd.data_frame.wrap_external_to_pandas_transformation(
                        DataFrame.toPandas,
                        get_inputs_lineage_fn
                    )
                    logging.info('KENSU: done patching DataFrame.toPandas')
                except:
                    import traceback
                    logging.warning("KENSU: unexpected issue when patching DataFrame.toPandas: {}".format(traceback.format_exc()))
                try:
                    logging.info('KENSU: patching spark.createDataFrame to work with pandas dataframes')
                    from pyspark.sql import SparkSession
                    SparkSession.createDataFrame = patched_spark_createDataFrame_pandas(
                        SparkSession.createDataFrame,
                        pandas_to_spark_df_via_tmp_file=pandas_to_spark_df_via_tmp_file,
                        tmp_dir=pandas_to_spark_tmp_dir
                    )
                    logging.info('KENSU: patching  spark.createDataFrame done')
                except:
                    import traceback
                    logging.warning("KENSU: unexpected issue when patching DataFrame.createDataFrame: {}".format(traceback.format_exc()))
                try:
                    logging.info('KENSU: adding spark env var KSU_DISABLE_PY_COLLECTOR=true to disable kensu-py collector on executor nodes')
                    spark_session.conf.set("KSU_DISABLE_PY_COLLECTOR", 'true')
                    spark_session.sparkContext.environment['KSU_DISABLE_PY_COLLECTOR'] = 'true'
                    logging.info('KENSU: done adding KSU_DISABLE_PY_COLLECTOR')
                except:
                    import traceback
                    logging.warning("KENSU: unexpected issue when adding spark.executorEnv.KSU_DISABLE_PY_COLLECTOR: {}".format(traceback.format_exc()))

            if use_api_client:
                from kensu.utils.kensu_provider import KensuProvider as K
                K().initKensu(do_report=False)

            if h2o_support:
                # needed to support both ways of importing: from local file or dam-client-python package
                try:
                    from kensu_dam.h2o import dam_patch_h2o
                except ImportError:
                    from dam_h2o_collector import dam_patch_h2o

                if logs_dir_path is not None:
                    dam_patch_h2o(log_file_prefix=process_name or '',
                                  debug_file_enabled=True,
                                  debug_stdout_enabled=True,
                                  dam_logs_directory=logs_dir_path,
                                  create_virtual_training_datasource=h2o_create_virtual_training_datasource)
                else:
                    dam_patch_h2o(debug_file_enabled=False,
                                  debug_stdout_enabled=True,
                                  create_virtual_training_datasource=h2o_create_virtual_training_datasource)

        except Exception as e:
            import traceback
            logging.warning("Error when initializing Kensu tracking: " + traceback.format_exc())
    else:
        logging.info("Tracking by Kensu is disabled")


def check_spark_circuit_breakers_and_stop_if_broken(
        spark,
        stop_spark=True):
    cls = ref_scala_object(spark.sparkContext._jvm, "io.kensu.sparkcollector.KensuSparkCollector")
    breakers_failed = cls.checkIfCircuitBreakersFailedAndPrepShutdown(spark._jsparkSession)
    if breakers_failed:
        logging.warning("Some kensu circuit breaker has failed (there are active unresolved tickets in a rule "
                        "marked as a circuit breaker), so the app will shutdown now")
        if stop_spark:
            logging.warning("Shutting down Spark context now...")
            spark.stop()
        else:
            logging.warning("Not shutting down Spark context as it was not requested.")
    return breakers_failed


def is_databricks():
    import os
    return "DATABRICKS_RUNTIME_VERSION" in os.environ


def log_and_print(log_fn, msg):
    log_fn(msg)
    # in Databricks print to stdout too, so output is visible in the same notebook
    # (default logging loglevel is WARNING, so logging.info would be lost)
    if is_databricks():
        print(msg)


def with_catch_errors(fn_name, default_result, fn_result_lambda):
    # TODO: check configurable decorators too https://stackoverflow.com/a/27446895 ?
    result = default_result
    try:
        logging.info('KENSU: in {}'.format(fn_name))
        result = fn_result_lambda(default_result)
        logging.info('KENSU: {} done'.format(fn_name))
    except:
        import traceback
        log_and_print(logging.warning, "KENSU: unexpected issue in {}: {}".format(
            fn_name, traceback.format_exc()
        ))
    return result


def catch_errors_with_default_self(func):
    def new_func(*args, **kwargs):
        default_self = args[0]
        return with_catch_errors(
            fn_name=func.__name__,
            default_result=default_self,
            fn_result_lambda=lambda default_result: func(*args, **kwargs))
    new_func.__doc__ = func.__doc__
    new_func.__name__ = func.__name__
    try:
        import inspect
        new_func.__signature__ = inspect.signature(func)
    except:
        pass
    return new_func


@catch_errors_with_default_self
def addOutputObservations(df,  # type: DataFrame
                          compute_count_distinct=False  # not recommended due to likely performance impact
                          ):
    # FIXME: looks like we have a repeating pyspark DF -> JVM DF -> pyspark DF pattern here
    spark = df.sql_ctx.sparkSession
    jvm = spark.sparkContext._jvm
    cls = ref_scala_object(jvm, "org.apache.spark.sql.kensu.KensuObserveMetrics")
    #   def addOutputObservations(df: DataFrame, computeCountDistinct: Boolean = false): DataFrame
    jdf = cls.addOutputObservations(df._jdf, compute_count_distinct)
    # finally convert Java DataFrame back to python DataFrame
    from pyspark.sql.dataframe import DataFrame
    return DataFrame(jdf, spark)


@catch_errors_with_default_self
def addOutputObservationsWithRemoteConf(df,  # type: DataFrame
                                        path=None,  # type: str
                                        table_name=None,  # type: str
                                        jdbc_options=None, # type: Optional[Dict[str, str]]
                                        format=None,
                                        compute_count_distinct=False  # not recommended due to likely performance impact
                                        ):
    # FIXME: looks like we have a repeating pyspark DF -> JVM DF -> pyspark DF pattern here
    spark = df.sql_ctx.sparkSession
    jvm = spark.sparkContext._jvm
    cls = ref_scala_object(jvm, "org.apache.spark.sql.kensu.KensuObserveMetrics")
    #   addOutputObservationsWithRemoteConf(df: DataFrame,
    #                                           maybeDsPath: String,
    #                                           maybeTableName: String,
    #                                           maybeJdbcOptions: java.util.HashMap[String, String],
    #                                           format: String,
    #                                           computeCountDistinct: Boolean
    #                                          )
    jdf = cls.addOutputObservationsWithRemoteConf(df._jdf, path, table_name, jdbc_options or {}, format, compute_count_distinct)
    # finally convert Java DataFrame back to python DataFrame
    from pyspark.sql.dataframe import DataFrame
    return DataFrame(jdf, spark)


# a global var, used in addCustomObservationsToOutput
next_observation_num = 0


@catch_errors_with_default_self
def addCustomObservationsToOutput(df,  # type: DataFrame
                                  *exprs  # type: pyspark.sql.column.Column
                                 ):
    """
    custom observations/metrics can be added at any stage of the stream DataFrame - and currently uses same semantics as
     pyspark's own DataFrame.observe(), so can pass any Spark aggregation expressions.

    just make sure that each expression:
    - return a double datatype .cast("double")
    - has a unique and meaningful alias within the whole stream DataFrame
      because that will become metrics name in Kensu, e.g. expr.alias("someColumn.max_before_filtering")

    see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrame.observe.html
    """
    global next_observation_num
    next_observation_num = next_observation_num + 1
    return df.observe(
        # observation ID must be unique and start with ksu_metrics_output_
        "ksu_metrics_output_custom_gen{}".format(next_observation_num),
       *exprs
    )


def make_kensu_efficient_write_fn(kensu_efficient_write_compute_count_distinct=False):
    from pyspark.sql.dataframe import DataFrameWriter, DataFrame

    def kensu_efficient_write(self):
        """Returns KensuDataFrameWriter,
        which is mostly same as Spark DataFrame.write, but adds Kensu Observations via .observe()"""
        from kensu.pyspark.KensuDataFrameWriter import KensuDataFrameWriter
        df_writer = KensuDataFrameWriter(self, compute_count_distinct=kensu_efficient_write_compute_count_distinct)
        return df_writer

    @property
    def kensu_efficient_write_property(self):
        """Same as Spark DataFrame.write but adds efficient Kensu Observations"""
        return kensu_efficient_write(self)

    return kensu_efficient_write_property


def pyspark_datatype_to_jvm(jvm, pyspark_datatype):
    return ref_scala_object(jvm, "org.apache.spark.sql.types.DataType")\
        .fromJson(pyspark_datatype.json())


@catch_errors_with_default_self
def reportCustomKafkaInputSchema(df,  # type: DataFrame
                                 custom_schema=None,  # StructType, if None, will use current DataFrame schema
                                 fieldNamePrefix="value.",
                                 schemaDesc="after Kafka event value parsing"
                                ):
    """
    this reports a custom schema for a Kafka read operations in Spark Streaming.
    it works by finding all/any such source URIs in the provided `SparkStreaming DataFrame`,
    and for each (if any), reporting the provided custom schema

    P.S. custom schema may appear in Kensu a few minutes later if the source was not yet ingested before (ingestion delay)
    """
    jvm = get_jvm_from_df(df)
    cls = ref_scala_object(jvm, "org.apache.spark.sql.kensu.KafkaStreamCustomSchema")
    jvm_datatype = pyspark_datatype_to_jvm(jvm, custom_schema or df.schema())
    #   def reportCustomSchema(df: DataFrame, custom_schema: StructType, fieldNamePrefix: String, schemaDesc: String): DataFrame
    res = cls.reportCustomSchema(df._jdf, jvm_datatype, fieldNamePrefix, schemaDesc)
    log_and_print(logging.info, 'Tried reporting custom Kafka schema: {}'.format(res))
    return df


@catch_errors_with_default_self
def reportCustomEventhubInputSchema(df,  # type: DataFrame
                                 custom_schema=None,  # StructType, if None, will use current DataFrame schema
                                 fieldNamePrefix="body.",
                                 schemaDesc="after EventHub event value parsing"
                                 ):
    """
    this reports a custom schema for a EventHub read operations in Spark Streaming.
    it works by finding all/any such source URIs in the provided `SparkStreaming DataFrame`,
    and for each (if any), reporting the provided custom schema

    P.S. custom schema may appear in Kensu a few minutes later if the source was not yet ingested before (ingestion delay)
    """
    jvm = get_jvm_from_df(df)
    cls = ref_scala_object(jvm, "org.apache.spark.sql.kensu.EventHubsCustomSchema")
    jvm_datatype = pyspark_datatype_to_jvm(jvm, custom_schema or df.schema())
    #   def reportCustomSchema(df: DataFrame, custom_schema: StructType, fieldNamePrefix: String, schemaDesc: String): DataFrame
    res = cls.reportCustomSchema(df._jdf, jvm_datatype, fieldNamePrefix, schemaDesc)
    log_and_print(logging.info, 'Tried reporting custom EventHub schema: {}'.format(res))
    return df



from matplotlib.pyplot import *
import matplotlib.pyplot as plt
from functools import partial

from kensu.utils.helpers import eventually_report_in_mem, get_absolute_path


def subplots(*args,**kwargs):
    orig_return = plt.subplots(*args, **kwargs)

    for i in range(len(orig_return[1])):
        def add_inheritance(self, arg):
            self.inheritance.append(arg)
        orig_return[1][i].inheritance = []
        orig_return[1][i].add_inheritance = partial(add_inheritance, orig_return[1][i])

        from kensu.utils.wrappers import kensu_wrapper_and_report
        for f in dir(orig_return[1][i]):
            if f in ['bar','boxplot','hist','hist2d','plot','plot_date']:
                exec("setattr(orig_return[1][i],f, kensu_wrapper_and_report(orig_return[1][i].%s,orig_return[1][i]))" % f)
    return orig_return



def savefig(*args, **kwargs):
    fig = gcf()
    res = fig.savefig(*args, **kwargs)
    fig.canvas.draw_idle() # need this if 'transparent=True' to reset colors

    from kensu.utils.kensu_provider import KensuProvider
    kensu = KensuProvider().instance()

    name = args[0]
    location = get_absolute_path(args[0])

    fig_ds = kensu.extractors.extract_data_source(fig, kensu.default_physical_location_ref,
                                                  logical_data_source_naming_strategy=kensu.logical_data_source_naming_strategy, location=location)._report()
    fig_sc = kensu.extractors.extract_schema(fig_ds,fig)._report()

    kensu.real_schema_df[fig_sc.to_guid()] = fig

    for ax in fig.axes:
        if (ax.inheritance) is not []:
            for element in ax.inheritance:
                orig_ds = eventually_report_in_mem(
                    kensu.extractors.extract_data_source(element, kensu.default_physical_location_ref,
                                                         logical_data_source_naming_strategy=kensu.logical_data_source_naming_strategy))

                orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, element))

                for col in [k.name for k in orig_sc.pk.fields]:
                    kensu.add_dependencies_mapping(fig_sc.to_guid(), col, orig_sc.to_guid(), col, 'fig')
    kensu.report_with_mapping()

    return res


import gluonts.mx.model.predictor as pred

from kensu.gluonts.ksu_utils.dataset_helpers import make_dataset_reliable
from kensu.utils.helpers import eventually_report_in_mem
from gluonts.dataset.common import ListDataset
from kensu.utils.kensu_provider import KensuProvider
from kensu.gluonts.model.forecast import SampleForecast

class RepresentableBlockPredictor(pred.RepresentableBlockPredictor):

    def predict(self, Y, *args, **kwargs):
        Y, old_Field, dep_fields = make_dataset_reliable(Y)

        original_result = list(super(RepresentableBlockPredictor, self).predict(dataset=Y, *args, **kwargs))


        if isinstance(Y, ListDataset):
            Y.list_data = old_Field

        deps = []
        kensu = KensuProvider().instance()

        for element in dep_fields:
            orig_ds = eventually_report_in_mem(
                kensu.extractors.extract_data_source(element, kensu.default_physical_location_ref,
                                                     logical_data_source_naming_strategy=kensu.logical_naming))
            orig_sc = eventually_report_in_mem(kensu.extractors.extract_schema(orig_ds, element))
            deps.append(orig_sc)

        def e(iterable):
            for b in iterable:
                b.__class__ = SampleForecast
                b.dependencies = deps
                yield b

        result = e(original_result)

        return result




from airflow.operators import python as airflow_python

from airflow.operators.python import *

from kensu.airflow.kensu_airflow_collector import COLLECTOR_STATUS_INIT, COLLECTOR_STATUS_DONE, log_status, \
    airflow_init_kensu, handle_ex


class PythonOperator(airflow_python.PythonOperator):
    def __init__(
            self,
            *args,
            **kwargs,
    ):
        super(PythonOperator, self).__init__(
            *args, **kwargs
        )

    def execute(self, context: Dict) -> None:
        # in PythonOperator, we need to initialize Kensu before calling the original execute(),
        # as the underlying python function could use Kensu tracked functions inside itself
        try:
            log_status(self, COLLECTOR_STATUS_INIT)
            airflow_init_kensu(airflow_operator=self)
        except Exception as ex:
            handle_ex(self, ex)
        res = super(PythonOperator, self).execute(context)
        log_status(self, COLLECTOR_STATUS_DONE)
        return res


PythonOperator.__doc__ = airflow_python.PythonOperator.__doc__
